


from ephyviewer.standalone import WindowManager, RawNeoOpenDialog
import pyqtgraph as pg



def test_WindowManager():
    app = pg.mkQApp()
    manager = WindowManager()
    manager.open_dialog()
    if manager.windows:
        app.exec_()


def test_RawNeoOpenDialog():
    app = pg.mkQApp()
    dia = RawNeoOpenDialog()
    dia.show()
    app.exec_()

    print(dia.final_params)



if __name__=='__main__':
    test_WindowManager()
    #~ test_RawNeoOpenDialog()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from collections import OrderedDict
import time
import sys
import pickle

from .myqt import QT, QT_MODE
from .navigation import NavigationToolBar

from .traceviewer import TraceViewer
from .epochviewer import EpochViewer
from .eventlist import EventList
from .spiketrainviewer import SpikeTrainViewer

location_to_qt={
    'left': QT.LeftDockWidgetArea,
    'right': QT.RightDockWidgetArea,
    'top': QT.TopDockWidgetArea,
    'bottom': QT.BottomDockWidgetArea,
}

orientation_to_qt={
    'horizontal': QT.Horizontal,
    'vertical': QT.Vertical,
}




class MainViewer(QT.QMainWindow):
    def __init__(self, debug=False, settings_name=None, parent=None, global_xsize_zoom=False, **navigation_params):
        QT.QMainWindow.__init__(self, parent)

        #TODO settings
        #http://www.programcreek.com/python/example/86789/PyQt5.QtCore.QSettings

        self.debug = debug
        if self.debug:
            print('debug True')
        self.settings_name = settings_name
        if self.settings_name is not None:
            pyver = '.'.join(str(e) for e in sys.version_info[0:3])
            appname = 'ephyviewer'+'_py'+pyver
            self.settings = QT.QSettings(appname, self.settings_name)
        self.global_xsize_zoom = global_xsize_zoom

        self.setDockNestingEnabled(True)


        self.viewers = OrderedDict()

        self.navigation_toolbar = NavigationToolBar(**navigation_params)

        dock = self.navigation_dock =  QT.QDockWidget('navigation',self)
        dock.setObjectName( 'navigation')
        dock.setWidget(self.navigation_toolbar)
        dock.setTitleBarWidget(QT.QWidget())
        self.addDockWidget(QT.TopDockWidgetArea, dock)

        self.navigation_toolbar.time_changed.connect(self.on_time_changed)
        self.navigation_toolbar.xsize_changed.connect(self.on_xsize_changed)
        self.navigation_toolbar.auto_scale_requested.connect(self.auto_scale)

        self.load_one_setting('navigation_toolbar', self.navigation_toolbar)


    def add_view(self, widget, location='bottom', orientation='vertical',
                        tabify_with=None, split_with=None):
        name = widget.name

        assert name not in self.viewers, 'Viewer already in MainViewer'

        dock = QT.QDockWidget(name)
        dock.setObjectName(name)
        dock.setWidget(widget)

        #TODO chustum titlebar
        #~ dock.setTitleBarWidget(titlebar)

        if tabify_with is not None:
            assert tabify_with in self.viewers, '{} no exists'.format(tabify_with)
            #~ raise(NotImplementedError)
            #tabifyDockWidget ( QDockWidget * first, QDockWidget * second )
            other_dock = self.viewers[tabify_with]['dock']
            self.tabifyDockWidget(other_dock, dock)

        elif split_with is not None:
            assert split_with in self.viewers, '{} no exists'.format(split_with)
            #~ raise(NotImplementedError)
            orien = orientation_to_qt[orientation]
            other_dock = self.viewers[split_with]['dock']
            self.splitDockWidget(other_dock, dock, orien)
            #splitDockWidget ( QDockWidget * first, QDockWidget * second, Qt::Orientation orientation )
        else:
            loc = location_to_qt[location]
            orien = orientation_to_qt[orientation]
            #~ print(location, loc)
            self.addDockWidget(loc, dock, orien)

        self.viewers[name] = {'widget': widget, 'dock':dock}

        self.load_one_setting(name, widget)

        widget.time_changed.connect(self.on_time_changed)
        if self.global_xsize_zoom and hasattr(widget, 'params_controller'):
            widget.params_controller.xsize_zoomed.connect(self.set_xsize)

        if hasattr(widget.source, 't_start'):
            # quick fix for DataFrameView should be removed with betetr solution
            if len(self.viewers) ==1:
                # first widget
                t_start, t_stop = widget.source.t_start, widget.source.t_stop
            else:
                    t_start = min(self.navigation_toolbar.t_start, widget.source.t_start)
                    t_stop = max(self.navigation_toolbar.t_stop, widget.source.t_stop)
            self.navigation_toolbar.set_start_stop(t_start, t_stop, seek=True)



    def load_one_setting(self, name, widget):
        #~ print('load_one_setting', name, self.settings_name)
        if self.settings_name is not None:
            value = self.settings.value('viewer_'+name)
            #~ print('value', value)
            if value is not None:
                try:
                #~ if True:
                    if QT_MODE == 'PyQt4' and sys.version_info[0]==2:
                        if type(value)==QT.QVariant:
                            value = bytes(value.toPyObject())
                    value = pickle.loads(value)
                    widget.set_settings(value)
                except:
                    print('erreur load settings', name)


    def save_all_settings(self):
        if self.debug:
            print('save_all_settings')
        if self.settings_name is not None:
            for name, d in self.viewers.items():
                value = d['widget'].get_settings()
                #~ print('save', name, type(value))
                if value is not None:
                    #~ print('save ', name)
                    self.settings.setValue('viewer_'+name, pickle.dumps(value))
            value = self.navigation_toolbar.get_settings()
            if value is not None:
                #~ print('save ', 'navigation_toolbar')
                self.settings.setValue('viewer_navigation_toolbar', pickle.dumps(value))

    def on_time_changed(self, t):

        for name , viewer in self.viewers.items():
            if viewer['widget'] != self.sender():
                t0 = time.time()
                viewer['widget'].seek(t)

                if self.debug:
                    t1 = time.time()
                    print('refresh duration for', name, t1-t0, 's')

        if self.navigation_toolbar != self.sender():
            #~ print('self.navigation_toolbar.seek', t)
            self.navigation_toolbar.seek(t, emit=False)

    def on_xsize_changed(self, xsize):
        #~ print('on_xsize_changed', xsize)
        for name , viewer in self.viewers.items():
            if hasattr(viewer['widget'], 'set_xsize'):
                viewer['widget'].set_xsize(xsize)

    def auto_scale(self):
        #~ print('on_xsize_changed', xsize)
        for name , viewer in self.viewers.items():
            if hasattr(viewer['widget'], 'auto_scale'):
                viewer['widget'].auto_scale()


    def seek(self, t):
        for name , viewer in self.viewers.items():
            viewer['widget'].seek(t)

        self.navigation_toolbar.seek(t, emit=False)

    def set_xsize(self, xsize):
        if hasattr(self.navigation_toolbar, 'spinbox_xsize'):
            self.navigation_toolbar.spinbox_xsize.setValue(xsize)
        #~ self.on_xsize_changed(xsize)

    def closeEvent(self, event):
        for name , viewer in self.viewers.items():
            viewer['widget'].close()
        self.save_all_settings()
        event.accept()





def compose_mainviewer_from_sources(sources, mainviewer=None):
    """
    Helper that compose a windows from several source with basic rules.

    Use internally in:
      * standalone
      * when generating mainviewer from neo segment

    """

    if mainviewer is None:
        mainviewer = MainViewer(show_auto_scale=True)

    for i, sig_source in enumerate(sources['signal']):
        view = TraceViewer(source=sig_source, name='signal {}'.format(i))
        view.params['scale_mode'] = 'same_for_all'
        view.params['display_labels'] = True
        if i==0:
            mainviewer.add_view(view)
        else:
            mainviewer.add_view(view, tabify_with='signal {}'.format(i-1))
        view.auto_scale()


    for i, spike_source in enumerate(sources['spike']):
        view = SpikeTrainViewer(source=spike_source, name='spikes')
        mainviewer.add_view(view)

    for i, ep_source in enumerate(sources['epoch']):
        view = EpochViewer(source=ep_source, name='epochs')
        mainviewer.add_view(view)

    if 'event' in sources and len(sources['event']) > 0:
        ev_source_list = sources['event']
    else:
        ev_source_list = sources['epoch']
    for i, ev_source in enumerate(ev_source_list):
        view = EventList(source=ev_source, name='Event list')
        mainviewer.add_view(view, location='bottom',  orientation='horizontal')


    return mainviewer



import ephyviewer


from  ephyviewer.tests.testing_tools import make_fake_video_source


def test_videoviewer():
    source = make_fake_video_source()

    #~ exit()
    #~ print('*'*50)
    app = ephyviewer.mkQApp()
    view = ephyviewer.VideoViewer(source=source, name='video')
    view.params['nb_column'] = 2

    win = ephyviewer.MainViewer(debug=True)
    win.add_view(view)
    win.show()

    app.exec_()



if __name__=='__main__':
    test_videoviewer()



import ephyviewer

import numpy as np
import pandas as pd



def test_dataframe_view():


    df = pd.DataFrame()
    df['time'] = np.arange(0,10,.25)
    df['factorA'] = ['a', 'b', 'c', 'd']*10
    df['factorB'] = [10,20]*20
    df['quality'] = ['good', 'bad']*20


    app = ephyviewer.mkQApp()
    view = ephyviewer.DataFrameView(source=df, name='table')

    win = ephyviewer.MainViewer(debug=True)
    win.add_view(view)
    win.show()

    app.exec_()



if __name__=='__main__':
    test_dataframe_view()



 
import sys
from ephyviewer import QT, mkQApp, QT_MODE
print('QT_MODE', QT_MODE)



if __name__ == '__main__' :
	app = mkQApp()
	
	w = QT.QWidget()
	w.show()
	w.setWindowIcon(QT.QIcon(':/media-playback-start.svg'))
	
	app.exec_()
	
	



# -*- coding: utf-8 -*-
"""

Helper for import PyQt5/PtQt4
see
http://mikeboers.com/blog/2015/07/04/static-libraries-in-a-dynamic-world#the-fold
"""


class ModuleProxy(object):

    def __init__(self, prefixes, modules):
        self.prefixes = prefixes
        self.modules = modules

    def __getattr__(self, name):
        for prefix in self.prefixes:
            fullname = prefix + name
            for module in self.modules:
                obj = getattr(module, fullname, None)
                if obj is not None:
                    setattr(self, name, obj) # cache it
                    return obj
        raise AttributeError(name)

QT_MODE = None
try:
    import PyQt5
    from PyQt5 import QtCore, QtGui, QtWidgets
    QT_MODE = 'PyQt5'
except :
    try:
        import PyQt4
        from PyQt4 import QtCore, QtGui
        QT_MODE = 'PyQt4'
    except ImportError:
        print('no PyQt5/PyQt4')
#~ print(QT_MODE)

if QT_MODE == 'PyQt5':
    #~ from PyQt5 import QtCore, QtGui, QtWidgets
    QT = ModuleProxy(['', 'Q', 'Qt'], [QtCore.Qt, QtCore, QtGui, QtWidgets])
elif QT_MODE == 'PyQt4':
    #~ from PyQt4 import QtCore, QtGui
    QT = ModuleProxy(['', 'Q', 'Qt'], [QtCore.Qt, QtCore, QtGui])
else:
    QT = None

if QT is not None:
    from pyqtgraph import mkQApp



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from copy import deepcopy
from collections import deque

import numpy as np

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
from . import tools

import pyqtgraph as pg

from .base import ViewerBase, Base_ParamController, MyViewBox, same_param_tree
from .epochviewer import RectItem, DataGrabber
from .datasource import WritableEpochSource


default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 3., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'label_fill_color', 'type': 'color', 'value': '#222222DD'},
    {'name': 'label_size', 'type': 'int', 'value': 8, 'limits': (1,np.inf)},
    {'name': 'new_epoch_step', 'type': 'float', 'value': .1, 'step': 0.1, 'limits':(0,np.inf)},
    {'name': 'exclusive_mode', 'type': 'bool', 'value': True},
    {'name': 'view_mode', 'type': 'list', 'value':'stacked', 'values' : ['stacked', 'flat']},
    {'name': 'keys_as_ticks', 'type': 'bool', 'value': True},
    {'name': 'undo_history_size', 'type': 'int', 'value': 500, 'limits': (1, np.inf)},

    #~ {'name': 'display_labels', 'type': 'bool', 'value': True},
    ]

SEEK_COL = 0
START_COL = 1
STOP_COL = 2
DURATION_COL = 3
LABEL_COL = 4
SPLIT_COL = 5
DUPLICATE_COL = 6
DELETE_COL = 7



class EpochEncoder_ParamController(Base_ParamController):
    def __init__(self, parent=None, viewer=None, with_visible=True, with_color=True):
        Base_ParamController.__init__(self, parent=parent, viewer=viewer)

        self.resize(400, 700)

        h = QT.QHBoxLayout()
        self.mainlayout.addLayout(h)

        self.v1 = QT.QVBoxLayout()
        h.addLayout(self.v1)
        self.tree_params = pg.parametertree.ParameterTree()
        self.tree_params.setParameters(self.viewer.params, showTop=True)
        self.tree_params.header().hide()
        self.v1.addWidget(self.tree_params, stretch = 1)

        self.tree_label_params = pg.parametertree.ParameterTree()
        self.tree_label_params.setParameters(self.viewer.by_label_params, showTop=True)
        self.tree_label_params.header().hide()
        self.v1.addWidget(self.tree_label_params, stretch = 3)

        self.btn_new_label = QT.PushButton('New label')
        self.btn_new_label.clicked.connect(self.create_new_label)
        self.v1.addWidget(self.btn_new_label)

    def create_new_label(self):
        label, ok = QT.QInputDialog.getText(self, 'New label', 'Enter a new label:')

        # abort if user cancelled or typed nothing
        if not ok or not label: return

        # abort if label already exists
        if label in self.viewer.source.possible_labels: return

        # TODO: determine color from a color map?
        color = (120, 120, 120) # gray

        # add new label to WritableEpochSource
        self.viewer.source.possible_labels.append(label)
        self.viewer.source.color_labels.append(color)

        # get next unused shortcut key
        used_keys = [p['key'] for p in self.viewer.by_label_params]
        unused_keys = [k for k in '1234567890' if k not in used_keys]
        key = unused_keys[0] if len(unused_keys)>0 else ''

        # assign shortcuts without and with modifier key
        self.viewer.assign_label_shortcuts(label, key)

        # add new label to params
        name = 'label{}'.format(len(self.viewer.source.possible_labels)-1)
        children = [
            {'name': 'name', 'type': 'str', 'value': label, 'readonly':True},
            {'name': 'color', 'type': 'color', 'value': self.source.color_by_label(label)},
            {'name': 'key', 'type': 'str', 'value': key},
        ]
        self.viewer.by_label_params.addChild({'name': name, 'type': 'group', 'children': children})

        # clear and redraw plot to update labels and plot range
        self.viewer.plot.clear()
        self.viewer.initialize_plot()
        self.viewer.refresh()

        # add new label to range selector dropdown menu
        self.viewer.combo_labels.addItems([label])

        # refresh table to add new label to all label dropdown menus
        self.viewer.refresh_table()



class EpochEncoder(ViewerBase):
    _default_params = default_params

    request_data = QT.pyqtSignal(float, float, object)

    def __init__(self, **kargs):
        ViewerBase.__init__(self, **kargs)

        assert isinstance(self.source, WritableEpochSource)

        self.label_shortcuts = {}

        self.make_params()
        self.make_param_controller()
        self.set_layout()

        self.viewBox.doubleclicked.connect(self.show_params_controller)

        self.initialize_plot()

        self.on_range_visibility_changed(refresh=False)



        self.thread = QT.QThread(parent=self)
        self.datagrabber = DataGrabber(source=self.source)
        self.datagrabber.moveToThread(self.thread)
        self.thread.start()


        self.datagrabber.data_ready.connect(self.on_data_ready)
        self.request_data.connect(self.datagrabber.on_request_data)

        self.table_button_fixed_width = 32
        self.refresh_table()

        self.history = deque(maxlen=self.params['undo_history_size'])
        self.history_position = 0
        self.changes_since_save = 0
        self.append_history()

        # reverse the automatic increment of changes_since_save done by
        # append_history, then disable the save button
        self.changes_since_save = 0
        self.save_action.setEnabled(False)


    def make_params(self):
        # Create parameters
        self.params = pg.parametertree.Parameter.create(name='Global options',
                                                    type='group', children=self._default_params)
        self.params.param('xsize').setLimits((0, np.inf))


        keys = '1234567890'
        all = []
        for i, label in enumerate(self.source.possible_labels):
            # get string for shortcut key
            key = keys[i] if i<len(keys) else ''

            name = 'label{}'.format(i)
            children = [
                {'name': 'name', 'type': 'str', 'value': label, 'readonly':True},
                {'name': 'color', 'type': 'color', 'value': self.source.color_by_label(label)},
                {'name': 'key', 'type': 'str', 'value': key},
            ]
            all.append({'name': name, 'type': 'group', 'children': children})

            # assign shortcuts without and with modifier key
            self.assign_label_shortcuts(label, key)

        self.by_label_params = pg.parametertree.Parameter.create(name='Labels', type='group', children=all)


        self.all_params = pg.parametertree.Parameter.create(name='all param',
                                    type='group', children=[self.params, self.by_label_params])

        self.params.sigTreeStateChanged.connect(self.on_param_change)
        self.by_label_params.sigTreeStateChanged.connect(self.on_change_keys)


    def set_layout(self):
        # layout
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)

        self.viewBox = MyViewBox()

        self.graphicsview  = pg.GraphicsView()#useOpenGL = True)
        self.mainlayout.addWidget(self.graphicsview, 4)

        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.plot.hideButtons()
        self.graphicsview.setCentralItem(self.plot)

        self.controls = QT.QWidget()
        self.controls.setSizePolicy(QT.QSizePolicy.Preferred, QT.QSizePolicy.Fixed)
        self.mainlayout.addWidget(self.controls)

        h = QT.QHBoxLayout()
        h.setContentsMargins(5, 5, 0, 0)
        h.setSpacing(5)
        self.controls.setLayout(h)

        # Range box

        self.range_group_box = QT.QGroupBox('Time &range selector', checkable=True, checked=False)
        self.range_group_box.clicked.connect(self.on_range_visibility_changed)
        h.addWidget(self.range_group_box)

        range_group_box_layout = QT.QGridLayout()
        range_group_box_layout.setSpacing(0)
        self.range_group_box.setLayout(range_group_box_layout)

        range_shortcut = QT.QShortcut(self)
        range_shortcut.setKey('r')
        range_shortcut.activated.connect(self.range_group_toggle)
        self.range_group_box.setToolTip('Toggle with shortcut: r')

        spinboxs = []
        buts = []
        for i, but_text in enumerate(['Set start >', 'Set stop >']):
            but = QT.QPushButton(but_text)
            buts.append(but)
            range_group_box_layout.addWidget(but, i, 0)
            spinbox = pg.SpinBox(value=float(i), decimals = 8, bounds = (-np.inf, np.inf),step = 0.05, siPrefix=False, int=False)
            if 'compactHeight' in spinbox.opts:  # pyqtgraph >= 0.11.0
                spinbox.setOpts(compactHeight=False)
            range_group_box_layout.addWidget(spinbox, i, 1)
            spinbox.setSizePolicy(QT.QSizePolicy.Preferred, QT.QSizePolicy.Preferred, )
            spinbox.valueChanged.connect(self.on_spin_limit_changed)
            spinboxs.append(spinbox)
        self.spin_limit1, self.spin_limit2 = spinboxs
        buts[0].clicked.connect(self.set_limit1)
        buts[1].clicked.connect(self.set_limit2)

        limit1_shortcut = QT.QShortcut(self)
        limit1_shortcut.setKey('[')
        limit1_shortcut.activated.connect(buts[0].click)
        buts[0].setToolTip('Set start with shortcut: [')

        limit2_shortcut = QT.QShortcut(self)
        limit2_shortcut.setKey(']')
        limit2_shortcut.activated.connect(buts[1].click)
        buts[1].setToolTip('Set stop with shortcut: ]')

        self.combo_labels = QT.QComboBox()
        self.combo_labels.addItems(self.source.possible_labels)
        range_group_box_layout.addWidget(self.combo_labels, 2, 0, 1, 2)

        self.but_apply_region = QT.PushButton('Insert within range')
        range_group_box_layout.addWidget(self.but_apply_region, 3, 0, 1, 2)
        self.but_apply_region.clicked.connect(self.apply_region)
        self.but_apply_region.setToolTip('Insert with customizable shortcuts (see options)')

        self.but_del_region = QT.PushButton('Clear within range')
        range_group_box_layout.addWidget(self.but_del_region, 4, 0, 1, 2)
        self.but_del_region.clicked.connect(self.delete_region)

        # Table

        self.table_widget = QT.QTableWidget()
        h.addWidget(self.table_widget, stretch=1)
        self.table_widget.setSelectionMode(QT.QAbstractItemView.SingleSelection)
        self.table_widget.setSelectionBehavior(QT.QAbstractItemView.SelectRows)
        self.table_widget.cellClicked.connect(self.on_table_cell_click)
        self.table_widget.cellChanged.connect(self.on_table_cell_change)
        self.table_widget_icons = {
            'seek': QT.QIcon(':/epoch-encoder-seek.svg'),
            'split': QT.QIcon(':/epoch-encoder-split.svg'),
            'duplicate': QT.QIcon(':/epoch-encoder-duplicate.svg'),
            'delete': QT.QIcon(':/epoch-encoder-delete.svg'),
        }

        # Toolbar

        self.toolbar = QT.QToolBar()
        self.toolbar.setToolButtonStyle(QT.Qt.ToolButtonTextBesideIcon)
        self.toolbar.setIconSize(QT.QSize(16, 16))
        self.mainlayout.addWidget(self.toolbar)

        self.toggle_controls_visibility_action = self.toolbar.addAction('Hide controls', self.on_controls_visibility_changed)
        self.toggle_controls_visibility_button = self.toolbar.widgetForAction(self.toggle_controls_visibility_action)
        self.toggle_controls_visibility_button.setArrowType(QT.UpArrow)

        self.toolbar.addSeparator()

        self.save_action = self.toolbar.addAction('Save', self.on_save)
        self.save_action.setShortcut('Ctrl+s')  # automatically converted to Cmd+s on Mac
        self.save_action.setToolTip('Save with shortcut: Ctrl/Cmd+s')
        self.save_action.setIcon(QT.QIcon(':/epoch-encoder-save.svg'))

        self.undo_action = self.toolbar.addAction('Undo', self.on_undo)
        self.undo_action.setShortcut('Ctrl+z')  # automatically converted to Cmd+z on Mac
        self.undo_action.setToolTip('Undo with shortcut: Ctrl/Cmd+z')
        self.undo_action.setIcon(QT.QIcon(':/epoch-encoder-undo.svg'))

        self.redo_action = self.toolbar.addAction('Redo', self.on_redo)
        self.redo_action.setShortcut('Ctrl+y')  # automatically converted to Cmd+y on Mac
        self.redo_action.setToolTip('Redo with shortcut: Ctrl/Cmd+y')
        self.redo_action.setIcon(QT.QIcon(':/epoch-encoder-redo.svg'))

        self.toolbar.addAction('Options', self.show_params_controller)

        self.toolbar.addAction('New label', self.params_controller.create_new_label)

        self.toolbar.addAction('Merge neighbors', self.on_merge_neighbors)

        self.toolbar.addAction('Fill blank', self.on_fill_blank)

        self.allow_overlap_action = self.toolbar.addAction('Allow overlap')
        self.allow_overlap_action.setToolTip('Hold Shift when using shortcut keys to temporarily switch modes')
        self.allow_overlap_action.setCheckable(True)
        self.allow_overlap_action.setChecked(not self.params['exclusive_mode'])
        self.allow_overlap_action.toggled.connect(lambda checked: self.params.param('exclusive_mode').setValue(not checked))


    def make_param_controller(self):
        self.params_controller = EpochEncoder_ParamController(parent=self, viewer=self)
        self.params_controller.setWindowFlags(QT.Qt.Window)


    def closeEvent(self, event):

        if self.changes_since_save != 0:
            text = 'Do you want to save epoch encoder changes before closing?'
            title = 'Save?'
            mb = QT.QMessageBox.question(self, title,text,
                    QT.QMessageBox.Ok ,  QT.QMessageBox.Discard)
            if mb==QT.QMessageBox.Ok:
                self.source.save()

        self.thread.quit()
        self.thread.wait()
        event.accept()


    def initialize_plot(self):
        self.region = pg.LinearRegionItem(brush='#FF00FF20')
        self.region.setZValue(10)
        self.region.setRegion((self.spin_limit1.value(), self.spin_limit2.value()))
        self.plot.addItem(self.region, ignoreBounds=True)
        self.region.sigRegionChanged.connect(self.on_region_changed)

        self.vline = pg.InfiniteLine(angle=90, movable=False, pen=self.params['vline_color'])
        self.vline.setZValue(1) # ensure vline is above plot elements
        self.plot.addItem(self.vline)

        self.rect_items = []

        self.label_items = []
        for i, label in enumerate(self.source.possible_labels):
            color = self.by_label_params['label'+str(i), 'color']
            label_item = pg.TextItem(label, color=color, anchor=(0, 0.5), border=None, fill=self.params['label_fill_color'])
            label_item.setZValue(11)
            font = label_item.textItem.font()
            font.setPointSize(self.params['label_size'])
            label_item.setFont(font)
            self.plot.addItem(label_item)
            self.label_items.append(label_item)

        self.viewBox.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)


    def on_controls_visibility_changed(self):
        if self.controls.isVisible():
            self.controls.hide()
            self.toggle_controls_visibility_action.setText('Show controls')
            self.toggle_controls_visibility_button.setArrowType(QT.RightArrow)
        else:
            self.controls.show()
            self.toggle_controls_visibility_action.setText('Hide controls')
            self.toggle_controls_visibility_button.setArrowType(QT.UpArrow)

    def show_params_controller(self):
        self.params_controller.show()

    def on_param_change(self):
        self.allow_overlap_action.setChecked(not self.params['exclusive_mode'])
        self.vline.setPen(color=self.params['vline_color'])
        for label_item in self.label_items:
            font = label_item.textItem.font()
            font.setPointSize(self.params['label_size'])
            label_item.setFont(font)
        self.refresh()

        if self.params['undo_history_size'] != self.history.maxlen:
            while len(self.history) > self.params['undo_history_size']:
                if self.history_position > 0:
                    # preferentially remove old states
                    self.history.popleft()
                    self.history_position -= 1
                else:
                    # otherwise remove newer, undone states
                    self.history.pop()
            self.history = deque(self.history, maxlen=self.params['undo_history_size'])

            # if the saved state was bumped off either end of the history
            # queue, it will no longer be reachable via undo/redo, so set
            # changes_since_save to None
            if self.changes_since_save is not None:
                if not (0 <= self.history_position - self.changes_since_save < len(self.history)):
                    self.changes_since_save = None

            self.refresh_toolbar()

    def assign_label_shortcuts(self, label, key):

        if label not in self.label_shortcuts:
            # create new shortcuts
            shortcut_without_modifier = QT.QShortcut(self)
            shortcut_with_modifier    = QT.QShortcut(self)
            shortcut_without_modifier.activated.connect(lambda: self.on_label_shortcut(label, False))
            shortcut_with_modifier   .activated.connect(lambda: self.on_label_shortcut(label, True))
            self.label_shortcuts[label] = (shortcut_without_modifier, shortcut_with_modifier)
        else:
            # get existing shortcuts
            shortcut_without_modifier, shortcut_with_modifier = self.label_shortcuts[label]

        # set/change the shortcut keys
        shortcut_without_modifier.setKey(key)
        shortcut_with_modifier   .setKey('Shift+' + key)

    def on_change_keys(self, refresh=True):

        for i, label in enumerate(self.source.possible_labels):
            # get string for shortcut key
            key = self.by_label_params['label'+str(i), 'key']

            # assign shortcuts without and with modifier key
            self.assign_label_shortcuts(label, key)

        self.refresh()

    def set_xsize(self, xsize):
        self.params['xsize'] = xsize


    def set_settings(self, value):
        #~ print('set_settings')
        actual_value = self.all_params.saveState()
        #~ print('same tree', same_param_tree(actual_value, value))
        if same_param_tree(actual_value, value):
            # this prevent restore something that is not same tree
            # as actual. Possible when new features.
            self.params.blockSignals(True)
            self.by_label_params.blockSignals(True)

            self.all_params.restoreState(value)

            self.params.blockSignals(False)
            self.by_label_params.blockSignals(False)
            self.on_change_keys(refresh=False)
        else:
            print('Not possible to restore setiings')

    def get_settings(self):
        return self.all_params.saveState()

    def refresh(self):
        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
        self.request_data.emit(t_start, t_stop, [0])

    def on_data_ready(self, t_start, t_stop, visibles, data):
        #~ print('on_data_ready', self, t_start, t_stop, visibles, data)
        #~ self.plot.clear()

        for rect_item in self.rect_items:
            self.plot.removeItem(rect_item)
        self.rect_items = []

        self.graphicsview.setBackground(self.params['background_color'])

        times, durations, labels, ids = data[0]
        #~ print(data)
        n = len(self.source.possible_labels)

        for i, label in enumerate(labels):
            ind = self.source.possible_labels.index(label)
            color = self.by_label_params['label'+str(ind), 'color']
            color2 = QT.QColor(color)
            color2.setAlpha(130)
            if self.params['view_mode']=='stacked':
                ypos = n-ind-1
            else:
                ypos = 0
            item = RectItem([times[i], ypos, durations[i], .9], border=color, fill=color2, id=ids[i])
            item.clicked.connect(self.on_rect_clicked)
            item.doubleclicked.connect(self.on_rect_doubleclicked)
            item.setPos(times[i],  ypos)
            self.plot.addItem(item)
            self.rect_items.append(item)


        ticks = []
        for i, label_item in enumerate(self.label_items):
            if self.params['view_mode']=='stacked':
                color = self.by_label_params['label'+str(i), 'color']
                #~ label_item = pg.TextItem(label, color=color, anchor=(0, 0.5), border=None, fill=pg.mkColor((128,128,128, 120)))
                label_item.setColor(color)
                label_item.fill = pg.mkBrush(self.params['label_fill_color'])
                ypos = n-i-0.55
                label_item.setPos(t_start, ypos)
                ticks.append((ypos, self.by_label_params['label'+str(i), 'key']))
                label_item.show()
                #~ self.plot.addItem(label_item)
            else:
                label_item.hide()

        if self.params['keys_as_ticks'] and self.params['view_mode']=='stacked':
            self.plot.getAxis('left').setTicks([ticks, []])
        else:
            self.plot.getAxis('left').setTicks([])

        if self.range_group_box.isChecked():
            self.region.show()
        else:
            self.region.hide()

        self.vline.setPos(self.t)
        self.plot.setXRange( t_start, t_stop, padding = 0.0)
        if self.params['view_mode']=='stacked':
            self.plot.setYRange( 0, n)
        else:
            self.plot.setYRange( 0, 1)

    def on_label_shortcut(self, label, modifier_used):

        range_selection_is_enabled = self.range_group_box.isChecked()

        if range_selection_is_enabled:
            # use selection for start and end of new epoch
            t_start = self.spin_limit1.value()
            t_stop = self.spin_limit2.value()
            duration = t_stop - t_start
        else:
            # use current time and step size to get end of new epoch
            duration = self.params['new_epoch_step']
            t_start = self.t
            t_stop = self.t + duration

        # delete existing epochs in the region where the new epoch will be inserted
        if (self.params['exclusive_mode'] and not modifier_used) or (not self.params['exclusive_mode'] and modifier_used):
            self.source.delete_in_between(t_start, t_stop)

        # create the new epoch
        self.source.add_epoch(t_start, duration, label)

        if not range_selection_is_enabled:
            # advance time by a step
            self.t += duration
            self.time_changed.emit(self.t)

        self.append_history()
        self.refresh()
        self.refresh_table()

    def on_merge_neighbors(self):
        self.source.merge_neighbors()
        self.append_history()
        self.refresh()
        self.refresh_table()

    def on_fill_blank(self):
        params = [{'name': 'method', 'type': 'list', 'value':'from_left', 'values' : ['from_left', 'from_right', 'from_nearest']}]
        dia = tools.ParamDialog(params, title='Fill blank method', parent=self)
        dia.resize(300, 100)
        if dia.exec_():
            d = dia.get()
            method = d['method']
            self.source.fill_blank(method=method)
            self.append_history()
            self.refresh()
            self.refresh_table()


    def on_save(self):
        self.source.save()
        self.changes_since_save = 0
        self.save_action.setEnabled(False)

    def append_history(self):
        # before appending a new state after the current position in the
        # history, remove any states that were undone
        while self.history_position < len(self.history) - 1:
            self.history.pop()

        if self.changes_since_save is not None:
            if self.changes_since_save < 0:
                # undone saved state was removed from history and so is
                # unreachable
                self.changes_since_save = None
            else:
                # saved state will be one more step in the past
                self.changes_since_save += 1

        self.history.append(deepcopy(self.source.all))
        self.history_position = len(self.history) - 1
        self.refresh_toolbar()

    def on_undo(self):
        if self.history_position > 0:
            self.history_position -= 1
            if self.changes_since_save is not None:
                self.changes_since_save -= 1
            self.source.all = deepcopy(self.history[self.history_position])
            self.refresh()
            self.refresh_table()
            self.refresh_toolbar()

    def on_redo(self):
        if self.history_position < len(self.history) - 1:
            self.history_position += 1
            if self.changes_since_save is not None:
                self.changes_since_save += 1
            self.source.all = deepcopy(self.history[self.history_position])
            self.refresh()
            self.refresh_table()
            self.refresh_toolbar()

    def refresh_toolbar(self):
        self.save_action.setEnabled(self.changes_since_save != 0)
        self.undo_action.setEnabled(self.history_position > 0)
        self.redo_action.setEnabled(self.history_position < len(self.history) - 1)

    def on_spin_limit_changed(self, v):
        # update region
        self.region.blockSignals(True)
        rgn = (self.spin_limit1.value(), self.spin_limit2.value())
        rgn = self.region.setRegion(rgn)
        self.region.blockSignals(False)

        # adjust spin box limits
        self.spin_limit1.setMaximum(self.spin_limit2.value())
        self.spin_limit2.setMinimum(self.spin_limit1.value())

    def on_region_changed(self):
        # update spin boxes
        self.spin_limit1.blockSignals(True)
        self.spin_limit2.blockSignals(True)
        rgn = self.region.getRegion()
        self.spin_limit1.setValue(rgn[0])
        self.spin_limit2.setValue(rgn[1])
        self.spin_limit1.blockSignals(False)
        self.spin_limit2.blockSignals(False)

        # adjust spin box limits
        self.spin_limit1.setMaximum(self.spin_limit2.value())
        self.spin_limit2.setMinimum(self.spin_limit1.value())

    def apply_region(self):

        rgn = self.region.getRegion()
        t = rgn[0]
        duration = rgn[1] - rgn[0]
        label = str(self.combo_labels.currentText())

        # delete existing epochs in the region where the new epoch will be inserted
        if self.params['exclusive_mode']:
            self.source.delete_in_between(rgn[0], rgn[1])

        # create the new epoch
        self.source.add_epoch(t, duration, label)

        self.append_history()
        self.refresh()
        self.refresh_table()

    def delete_region(self):

        rgn = self.region.getRegion()

        self.source.delete_in_between(rgn[0], rgn[1])

        self.append_history()
        self.refresh()
        self.refresh_table()

    def range_group_toggle(self):
        self.range_group_box.setChecked(not self.range_group_box.isChecked())
        self.on_range_visibility_changed()

    def on_range_visibility_changed(self, flag=None, refresh=True, shift_region=True):
        enabled = self.range_group_box.isChecked()

        for w in (self.spin_limit1, self.spin_limit2, self.combo_labels, self.but_apply_region, self.but_del_region):
            w.setEnabled(enabled)
        if enabled and shift_region:
            rgn = self.region.getRegion()
            rgn = (self.t, self.t + rgn[1] - rgn[0])
            self.region.setRegion(rgn)
        self.refresh()

    def set_limit1(self):
        if self.t<self.spin_limit2.value():
            self.spin_limit1.setValue(self.t)
            self.spin_limit1.repaint()  # needed on macOS

    def set_limit2(self):
        if self.t>self.spin_limit1.value():
            self.spin_limit2.setValue(self.t)
            self.spin_limit2.repaint()  # needed on macOS

    def refresh_table(self):
        self.table_widget.blockSignals(True)

        self.table_widget.clear()
        times, durations, labels = self.source.ep_times, self.source.ep_durations, self.source.ep_labels
        self.table_widget.setColumnCount(8)
        self.table_widget.setRowCount(times.size)
        self.table_widget.setHorizontalHeaderLabels(['', 'Start', 'Stop', 'Duration', 'Label', '', '', ''])

        # lock column widths for buttons to fixed button width
        self.table_widget.horizontalHeader().setMinimumSectionSize(self.table_button_fixed_width)
        for col in [SEEK_COL, SPLIT_COL, DUPLICATE_COL, DELETE_COL]:
            self.table_widget.horizontalHeader().setSectionResizeMode(col, QT.QHeaderView.Fixed)
            self.table_widget.setColumnWidth(col, self.table_button_fixed_width)

        for r in range(times.size):

            # seek button
            item = QT.QTableWidgetItem(self.table_widget_icons['seek'], '')
            item.setToolTip('Jump to epoch')
            self.table_widget.setItem(r, SEEK_COL, item)

            # start
            value = np.round(times[r], 6) # round to nearest microsecond
            item = QT.QTableWidgetItem('{}'.format(value))
            self.table_widget.setItem(r, START_COL, item)

            # stop
            value = np.round(times[r]+durations[r], 6) # round to nearest microsecond
            item = QT.QTableWidgetItem('{}'.format(value))
            self.table_widget.setItem(r, STOP_COL, item)

            # duration
            value = np.round(durations[r], 6) # round to nearest microsecond
            item = QT.QTableWidgetItem('{}'.format(value))
            self.table_widget.setItem(r, DURATION_COL, item)

            # label
            item = QT.QTableWidgetItem(labels[r])
            self.table_widget.setItem(r, LABEL_COL, item)

            # split button
            item = QT.QTableWidgetItem(self.table_widget_icons['split'], '')
            item.setToolTip('Split epoch at current time')
            self.table_widget.setItem(r, SPLIT_COL, item)

            # duplicate button
            item = QT.QTableWidgetItem(self.table_widget_icons['duplicate'], '')
            item.setToolTip('Duplicate epoch')
            self.table_widget.setItem(r, DUPLICATE_COL, item)

            # delete button
            item = QT.QTableWidgetItem(self.table_widget_icons['delete'], '')
            item.setToolTip('Delete epoch')
            self.table_widget.setItem(r, DELETE_COL, item)

        self.table_widget.blockSignals(False)

    def on_rect_clicked(self, id):

        # get index corresponding to epoch id
        ind = self.source.id_to_ind[id]

        # select the epoch in the data table
        self.table_widget.blockSignals(True)
        self.table_widget.setCurrentCell(ind, LABEL_COL) # select the label combo box
        self.table_widget.blockSignals(False)

    def on_rect_doubleclicked(self, id):

        # get index corresponding to epoch id
        ind = self.source.id_to_ind[id]

        # set region to epoch start and stop
        self.region.setRegion((self.source.ep_times[ind], self.source.ep_stops[ind]))

        # show the region if it isn't already visible
        self.range_group_box.setChecked(True)
        self.on_range_visibility_changed(shift_region=False)

    def on_seek_table(self, ind=None):
        if self.table_widget.rowCount()==0:
            return
        if ind is None:
            selected_ind = self.table_widget.selectedIndexes()
            if len(selected_ind)==0:
                return
            ind = selected_ind[0].row()
        self.t = self.source.ep_times[ind]
        self.refresh()
        self.time_changed.emit(self.t)

    def on_table_cell_click(self, row, col):
        if col == SEEK_COL:
            self.on_seek_table(row)
        elif col == SPLIT_COL:
            self.split_selected_epoch(row)
        elif col == DUPLICATE_COL:
            self.duplicate_selected_epoch(row)
        elif col == DELETE_COL:
            self.delete_selected_epoch(row)
        elif col == LABEL_COL:
            # replace text cell with a combo box
            combo_labels = QT.QComboBox()
            combo_labels.addItems(self.source.possible_labels)
            combo_labels.setCurrentText(self.source.ep_labels[row])
            combo_labels.currentIndexChanged.connect(
                lambda label_index, row=row: self.on_change_label(row, self.source.possible_labels[label_index])
            )
            self.table_widget.setCellWidget(row, LABEL_COL, combo_labels)
        else:
            return

    def on_table_cell_change(self, row, col):
        line_edit = self.table_widget.cellWidget(row, col)
        if not isinstance(line_edit, QT.QLineEdit): return
        new_text = line_edit.text()

        try:
            # convert string to number
            new_number = float(new_text)

        except ValueError:
            # an invalid number may have been entered, so restore the old value
            if col == START_COL:
                old_number = self.source.ep_times[row]
            elif col == STOP_COL:
                old_number = self.source.ep_times[row] + self.source.ep_durations[row]
            elif col == DURATION_COL:
                old_number = self.source.ep_durations[row]
            else:
                print('unexpected column changed')
                return
            old_number = np.round(old_number, 6) # round to nearest microsecond
            self.table_widget.blockSignals(True)
            self.table_widget.item(row, col).setText(str(old_number))
            self.table_widget.blockSignals(False)

        else:
            self.table_widget.blockSignals(True)

            # round and copy rounded number to table
            new_number = np.round(new_number, 6) # round to nearest microsecond
            self.table_widget.item(row, col).setText(str(new_number))

            if col == START_COL:

                # get the epoch stop time
                stop_time = self.source.ep_times[row] + self.source.ep_durations[row]

                # change epoch start time
                self.source.ep_times[row] = new_number

                # change epoch duration to keep stop time unchanged
                self.source.ep_durations[row] = stop_time - self.source.ep_times[row]

                # update epoch duration in table
                duration_line_edit = self.table_widget.item(row, DURATION_COL)
                new_duration = np.round(self.source.ep_durations[row], 6) # round to nearest microsecond
                duration_line_edit.setText(str(new_duration))

            elif col == STOP_COL:

                # change epoch duration for corresponding stop time
                self.source.ep_durations[row] = new_number-self.source.ep_times[row]

                # update epoch duration in table
                duration_line_edit = self.table_widget.item(row, DURATION_COL)
                new_duration = np.round(self.source.ep_durations[row], 6) # round to nearest microsecond
                duration_line_edit.setText(str(new_duration))

            elif col == DURATION_COL:

                # change epoch duration
                self.source.ep_durations[row] = new_number

                # update epoch stop time in table
                stop_line_edit = self.table_widget.item(row, STOP_COL)
                new_stop_time = self.source.ep_times[row] + self.source.ep_durations[row]
                new_stop_time = np.round(new_stop_time, 6) # round to nearest microsecond
                stop_line_edit.setText(str(new_stop_time))

            else:
                print('unexpected column changed')

            self.table_widget.blockSignals(False)

            self.append_history()

            # update plot
            self.refresh()
            # refresh_table is not called to avoid deselecting table cell

    def on_change_label(self, ind, new_label):

        # change epoch label
        self.source.ep_labels[ind] = new_label

        self.append_history()

        # update plot
        self.refresh()
        # refresh_table is not called to avoid deselecting table cell

    def delete_selected_epoch(self, ind=None):
        if self.table_widget.rowCount()==0:
            return
        if ind is None:
            selected_ind = self.table_widget.selectedIndexes()
            if len(selected_ind)==0:
                return
            ind = selected_ind[0].row()
        self.source.delete_epoch(ind)
        self.append_history()
        self.refresh()
        self.refresh_table()

    def duplicate_selected_epoch(self, ind=None):
        if self.table_widget.rowCount()==0:
            return
        if ind is None:
            selected_ind = self.table_widget.selectedIndexes()
            if len(selected_ind)==0:
                return
            ind = selected_ind[0].row()
        self.source.add_epoch(self.source.ep_times[ind], self.source.ep_durations[ind], self.source.ep_labels[ind])
        self.append_history()
        self.refresh()
        self.refresh_table()

    def split_selected_epoch(self, ind=None):
        if self.table_widget.rowCount()==0:
            return
        if ind is None:
            selected_ind = self.table_widget.selectedIndexes()
            if len(selected_ind)==0:
                return
            ind = selected_ind[0].row()
        if self.t <= self.source.ep_times[ind] or self.source.ep_stops[ind] <= self.t:
            return
        self.source.split_epoch(ind, self.t)
        self.append_history()
        self.refresh()
        self.refresh_table()



import os
import sys

from ephyviewer.myqt import QT_MODE
print('QT_MODE', QT_MODE)


def make_ressoureces_icons():
    with open('icons.qrc','w') as f:
        f.write("""<!DOCTYPE RCC><RCC version="1.0">
    <qresource>
""")
        for p, d, files in os.walk('./'):
            for filename in files:
                if filename.endswith('.png') or filename.endswith('.jpg') or filename.endswith('.svg'):
                    f.write('            <file alias="%s">%s/%s</file>\r\n' % (filename, p[2:],filename) )
                
        f.write("""    </qresource>
</RCC>
""")
    
    if QT_MODE=='PyQt4':
        if sys.version_info > (3,):
            os.popen('pyrcc4 -py3 icons.qrc -o icons_py3_Qt4.py')
        else:
            os.popen('pyrcc4 icons.qrc -o icons_py2_Qt4.py')
        
    elif QT_MODE=='PyQt5':
        if sys.version_info > (3,):
            os.popen('pyrcc5  icons.qrc -o icons_py3.py')
        else:
            raise(NotImplementedError)
    

#------------------------------------------------------------------------------
if __name__ == '__main__' :
    make_ressoureces_icons()




import ephyviewer


from  ephyviewer.tests.testing_tools import make_fake_epoch_source


def test_epoch_viewer():
    source = make_fake_epoch_source()


    app = ephyviewer.mkQApp()
    view = ephyviewer.EpochViewer(source=source, name='epoch')

    win = ephyviewer.MainViewer(debug=True)
    win.add_view(view)
    win.show()

    app.exec_()


if __name__=='__main__':
    test_epoch_viewer()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

from .sourcebase import BaseDataSource
from .events import BaseEventAndEpoch, InMemoryEventSource


class BaseSpikeSource(BaseEventAndEpoch):
    type = 'Spike'


class InMemorySpikeSource(BaseSpikeSource):

    def __init__(self, all_spikes=[]):
        BaseSpikeSource.__init__(self, all=all_spikes)

        s = [ np.max(e['time']) for e in self.all  if len(e['time'])>0]
        self._t_stop = max(s) if len(s)>0 else 0


    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        spike_times = self.all[chan]['time'][i_start:i_stop]
        return spike_times

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        spike_times = self.all[chan]['time']

        i1 = np.searchsorted(spike_times, t_start, side='left')
        i2 = np.searchsorted(spike_times, t_stop, side='left')
        sl = slice(i1, i2+1)
        return spike_times[sl]



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import ephyviewer
import os

import numpy as np
import matplotlib.pyplot as plt


def make_fake_signals():
    np.random.seed(2205)
    signals = np.random.randn(1000000, 16)

    signals *= np.random.rand(16)[None,:]
    signals += (np.random.rand(16)[None,:]-1)*3

    #~ print(signals.shape)
    sample_rate = 10000.
    t_start = 0.

    source = ephyviewer.InMemoryAnalogSignalSource(signals, sample_rate, t_start)


    return source


def make_fake_signals_with_scatter():
    sample_rate = 10000.
    t_start = 0.
    times = np.arange(1000000)/sample_rate
    signals = np.sin(times*2*np.pi*5)[:, None]

    s0 = signals[:-2, 0]
    s1 = signals[1:-1,0]
    s2 = signals[2:,0]

    peaks0,  = np.nonzero((s0<s1) & (s2<s1))
    peaks1,  = np.nonzero((s0>s1) & (s2>s1))

    scatter_indexes = {0: peaks0, 1: peaks1}
    scatter_channels = {0: [0], 1: [0]}
    source = ephyviewer.AnalogSignalSourceWithScatter(signals, sample_rate, t_start, scatter_indexes, scatter_channels)

    return source




def make_video_file(filename, codec='mpeg4', rate=25.): # mpeg4 mjpeg libx264
    print('make_video_file', filename)
    name=filename.replace('.avi', '')
    import av

    w, h = 800, 600

    output = av.open(filename, 'w')
    stream = output.add_stream(codec, rate)
    #~ stream.bit_rate = 8000000
    #~ stream.pix_fmt = 'yuv420p'
    stream.pix_fmt = 'yuv420p'
    stream.height = h
    stream.width = w

    duration = 10.


    #~ one_img = np.ones((h, w, 3), dtype='u1')
    #~ fig, ax = plt.subplots(figsize=(w/100, h/100), dpi=100,)
    fig = plt.figure(figsize=(w/100, h/100), dpi=100,)
    #~ line, = ax.plot([0], [0], marker='o', markersize=50, color='m')
    #~ ax.set_xlim(-1,1)
    #~ ax.set_ylim(-1,1)
    for i in range(int(duration*rate)):
        fig.clear()
        #~ one_img += np.random.randint(low=0, high=255, size=(h, w, 3)).astype('u1')
        text = '{} - frame {:5} - time {:6.3f}s'.format(name, i, i/rate)
        fig.text(.5, .5, text, ha='center', va='center')
        #~ ax.set_title()
        #~ line.set_markersize(i)
        fig.canvas.draw()
        one_img = np.fromstring(fig.canvas.tostring_rgb(), dtype='u1').reshape(h,w,3)
        one_img = one_img[:,:,::-1].copy()
        #~ one_img = one_img .swapaxes(0,1).copy()

        frame = av.VideoFrame.from_ndarray(one_img, format='bgr24')
        packet = stream.encode(frame)
        if packet is not None:
            output.mux(packet)

    while True :
        try:
            packet = stream.encode(None)
            if packet is None:
                break
            output.mux(packet)
        except av.AVError:
            break

    output.close()


def make_fake_video_source():
    filenames = ['video0.avi', 'video1.avi', 'video2.avi',]
    #~ filenames = ['video0.avi']
    for filename in filenames:
        if not os.path.exists(filename):
            make_video_file(filename, )



    source = ephyviewer.MultiVideoFileSource(filenames)

    return source



def make_fake_event_source():
    all_events = []
    for c in range(3):
        ev_times = np.arange(0, 10., .5) + c*3
        ev_labels = np.array(['Event{} num {}'.format(c, i) for i in range(ev_times.size)], dtype='U')
        all_events.append({ 'time':ev_times, 'label':ev_labels, 'name':'Event{}'.format(c) })

    source = ephyviewer.InMemoryEventSource(all_events=all_events)
    return source



def make_fake_epoch_source():
    all_epochs = []
    for c in range(3):
        ep_times = np.arange(0, 10., .5) + c*3
        ep_durations = np.ones(ep_times.shape) * .1
        ep_labels = np.array(['Event{} num {}'.format(c, i) for i in range(ep_times.size)], dtype='U')
        all_epochs.append({ 'time':ep_times, 'duration':ep_durations, 'label':ep_labels, 'name':'Event{}'.format(c) })

    source = ephyviewer.InMemoryEpochSource(all_epochs=all_epochs)
    return source


def make_fake_spiketrain_source():
    all_spikes =[]
    for c in range(30):
        spike_times = np.arange(0, 10., .5) + c*0.2
        all_spikes.append({ 'time':spike_times, 'name':'Unit#{}'.format(c) })

    source = ephyviewer.InMemorySpikeSource(all_spikes=all_spikes)
    return source



# -*- coding: utf-8 -*-
from .version import version as __version__

#common tools
from .myqt import *
from .icons import *
from .datasource import *
from .mainviewer import MainViewer, compose_mainviewer_from_sources
from .navigation import NavigationToolBar



#Viewers
from .traceviewer import TraceViewer
from .videoviewer import VideoViewer
from .eventlist import EventList
from .epochviewer import EpochViewer
from .timefreqviewer import TimeFreqViewer
from .dataframeview import DataFrameView
from .spiketrainviewer import SpikeTrainViewer


#Encoders
from .epochencoder import EpochEncoder



import ephyviewer


from  ephyviewer.tests.testing_tools import make_fake_event_source


def test_eventlist():
    source = make_fake_event_source()


    app = ephyviewer.mkQApp()
    view = ephyviewer.EventList(source=source, name='events')

    win = ephyviewer.MainViewer(debug=True)
    win.add_view(view)
    win.show()

    app.exec_()



if __name__=='__main__':
    test_eventlist()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from .myqt import QT
import pyqtgraph as pg
import numpy as np

import matplotlib.cm
import matplotlib.colors


import weakref




class ViewerBase(QT.QWidget):

    time_changed = QT.pyqtSignal(float)

    def __init__(self, name='', source=None, **kargs):
        QT.QWidget.__init__(self, **kargs)

        self.name = name
        self.source = source
        self.t = 0.

    def seek(self, t):
        self.t = t
        self.refresh()

    def refresh(self):
        #overwrite this one
        raise(NotImplementedError)

    def set_settings(self, value):
        pass

    def get_settings(self):
        return None

    def auto_scale(self):
        pass



class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    ygain_zoom = QT.pyqtSignal(float)
    xsize_zoom = QT.pyqtSignal(float)
    def __init__(self, *args, **kwds):
        pg.ViewBox.__init__(self, *args, **kwds)
        self.disableAutoRange()
    def mouseClickEvent(self, ev):
        if ev.double():
            ev.accept()
            self.doubleclicked.emit()
        else:
            ev.ignore()
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 5. if ev.delta()>0 else 1/5.
        else:
            z = 1.1 if ev.delta()>0 else 1/1.1
        self.ygain_zoom.emit(z)
        ev.accept()
    def mouseDragEvent(self, ev, axis=None):
        if ev.button()== QT.RightButton:
            self.xsize_zoom.emit((ev.pos()-ev.lastPos()).x())
        else:
            pass
        ev.accept()


class BaseMultiChannelViewer(ViewerBase):

    _default_params = None
    _default_by_channel_params = None
    _ControllerClass = None

    def __init__(self, with_user_dialog=True, **kargs):
        ViewerBase.__init__(self, **kargs)

        self.with_user_dialog = with_user_dialog

    def make_params(self):
        # Create parameters
        all = []
        for i in range(self.source.nb_channel):
            #TODO add name, hadrware index, id
            name = 'ch{}'.format(i)
            children =[{'name': 'name', 'type': 'str', 'value': self.source.get_channel_name(i), 'readonly':True}]
            children += self._default_by_channel_params
            all.append({'name': name, 'type': 'group', 'children': children})
        self.by_channel_params = pg.parametertree.Parameter.create(name='Channels', type='group', children=all)
        self.params = pg.parametertree.Parameter.create(name='Global options',
                                                    type='group', children=self._default_params)
        self.all_params = pg.parametertree.Parameter.create(name='all param',
                                    type='group', children=[self.params, self.by_channel_params])
        self.all_params.sigTreeStateChanged.connect(self.on_param_change)

    def set_layout(self, useOpenGL=None):
        # layout
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)

        self.viewBox = MyViewBox()

        self.graphicsview  = pg.GraphicsView(useOpenGL=useOpenGL)
        self.mainlayout.addWidget(self.graphicsview)

        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.plot.hideButtons()
        self.graphicsview.setCentralItem(self.plot)

    def make_param_controller(self):
        if self.with_user_dialog and self._ControllerClass:
            self.all_params.blockSignals(True)
            self.params_controller = self._ControllerClass(parent=self, viewer=self)
            self.params_controller.setWindowFlags(QT.Qt.Window)
            self.all_params.blockSignals(False)
        else:
            self.params_controller = None

    def show_params_controller(self):
        self.params_controller.show()

    def on_param_change(self):
        self.refresh()

    def set_xsize(self, xsize):
        #~ print(self.__class__.__name__, 'set_xsize', xsize)
        if 'xsize' in [p.name() for p in self.params.children()]:
            self.params['xsize'] = xsize

    def set_settings(self, value):
        actual_value = self.all_params.saveState()
        #~ print('same tree', same_param_tree(actual_value, value))
        if same_param_tree(actual_value, value):
            # this prevent restore something that is not same tree
            # as actual. Possible when new features.
            self.all_params.blockSignals(True)
            self.all_params.restoreState(value)
            self.all_params.blockSignals(False)
        else:
            print('Not possible to restore setiings')

    def get_settings(self):
        return self.all_params.saveState()


def same_param_tree(tree1, tree2):
    children1 = list(tree1['children'].keys())
    children2 = list(tree2['children'].keys())
    if len(children1) != len(children2):
        return False

    for k1, k2 in zip(children1, children2):
        if k1!=k2:
            return False

        if 'children' in tree1['children'][k1]:
            if not 'children' in tree2['children'][k2]:
                return False
            #~ print('*'*5)
            #~ print('Recursif', k1)
            if not same_param_tree(tree1['children'][k1], tree2['children'][k2]):
                return False

    return True


class Base_ParamController(QT.QWidget):

    xsize_zoomed = QT.pyqtSignal(float)

    def __init__(self, parent=None, viewer=None):
        QT.QWidget.__init__(self, parent)

        # this controller is an attribute of the viewer
        # so weakref to avoid loop reference and Qt crash
        self._viewer = weakref.ref(viewer)

        # layout
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)
        t = 'Options for {}'.format(self.viewer.name)
        self.setWindowTitle(t)
        self.mainlayout.addWidget(QT.QLabel('<b>'+t+'<\b>'))


    @property
    def viewer(self):
        return self._viewer()

    @property
    def source(self):
        return self._viewer().source

    def apply_xsize_zoom(self, xmove):
        MIN_XSIZE = 1e-6
        factor = xmove/100.
        factor = max(factor, -0.5)
        factor = min(factor, 1)
        newsize = self.viewer.params['xsize']*(factor+1.)
        self.viewer.params['xsize'] = max(newsize, MIN_XSIZE)
        self.xsize_zoomed.emit(self.viewer.params['xsize'])



class Base_MultiChannel_ParamController(Base_ParamController):
    channel_visibility_changed = QT.pyqtSignal()
    channel_color_changed = QT.pyqtSignal()

    def __init__(self, parent=None, viewer=None, with_visible=True, with_color=True):
        Base_ParamController.__init__(self, parent=parent, viewer=viewer)


        h = QT.QHBoxLayout()
        self.mainlayout.addLayout(h)

        self.v1 = QT.QVBoxLayout()
        h.addLayout(self.v1)
        self.tree_params = pg.parametertree.ParameterTree()
        self.tree_params.setParameters(self.viewer.params, showTop=True)
        self.tree_params.header().hide()
        self.v1.addWidget(self.tree_params)

        self.tree_by_channel_params = pg.parametertree.ParameterTree()
        self.tree_by_channel_params.header().hide()
        h.addWidget(self.tree_by_channel_params)
        self.tree_by_channel_params.setParameters(self.viewer.by_channel_params, showTop=True)
        v = QT.QVBoxLayout()
        h.addLayout(v)

        #~ but = QT.PushButton('default params')
        #~ v.addWidget(but)
        #~ but.clicked.connect(self.reset_to_default)

        if hasattr(self.viewer, 'auto_scale'):
            but = QT.PushButton('Auto scale')
            v.addWidget(but)
            but.clicked.connect(self.auto_scale_viewer)


        if with_visible:
            if self.source.nb_channel>1:
                v.addWidget(QT.QLabel('<b>Select channel...</b>'))
                names = [p.name() + ': '+p['name'] for p in self.viewer.by_channel_params]
                self.qlist = QT.QListWidget()
                v.addWidget(self.qlist, 2)
                self.qlist.addItems(names)
                self.qlist.setSelectionMode(QT.QAbstractItemView.ExtendedSelection)
                self.qlist.doubleClicked.connect(self.on_double_clicked)

                for i in range(len(names)):
                    self.qlist.item(i).setSelected(True)
                v.addWidget(QT.QLabel('<b>and apply...<\b>'))


            but = QT.QPushButton('set visble')
            v.addWidget(but)
            but.clicked.connect(self.on_set_visible)

            self.channel_visibility_changed.connect(self.on_channel_visibility_changed)

        if with_color:
            v.addWidget(QT.QLabel('<b>Set color<\b>'))
            h = QT.QHBoxLayout()
            but = QT.QPushButton('Progressive')
            but.clicked.connect(self.on_automatic_color)
            h.addWidget(but,4)
            self.combo_cmap = QT.QComboBox()
            self.combo_cmap.addItems(['Accent', 'Dark2','jet', 'prism', 'hsv', ])
            h.addWidget(self.combo_cmap,1)
            v.addLayout(h)

            self.channel_color_changed.connect(self.on_channel_color_changed)

    #~ def reset_to_default(self):
        #~ self.viewer.make_params()
        #~ self.tree_params.setParameters(self.viewer.params, showTop=True)
        #~ self.tree_by_channel_params.setParameters(self.viewer.by_channel_params, showTop=True)
        #~ ## self.viewer.on_param_change()
        #~ self.viewer.refresh()

    def auto_scale_viewer(self):
        self.viewer.auto_scale()

    @property
    def selected(self):
        selected = np.ones(self.viewer.source.nb_channel, dtype=bool)
        if self.viewer.source.nb_channel>1:
            selected[:] = False
            selected[[ind.row() for ind in self.qlist.selectedIndexes()]] = True
        return selected


    @property
    def visible_channels(self):
        visible = [self.viewer.by_channel_params['ch{}'.format(i), 'visible'] for i in range(self.source.nb_channel)]
        return np.array(visible, dtype='bool')

    def on_set_visible(self):
        # apply
        self.viewer.by_channel_params.blockSignals(True)

        visibles = self.selected
        for i,param in enumerate(self.viewer.by_channel_params.children()):
            param['visible'] = visibles[i]

        self.viewer.by_channel_params.blockSignals(False)
        self.channel_visibility_changed.emit()

    def on_double_clicked(self, index):
        self.viewer.by_channel_params.blockSignals(True)
        visibles = self.selected
        for i,param in enumerate(self.viewer.by_channel_params.children()):
            param['visible'] = (i==index.row())
        self.viewer.by_channel_params.blockSignals(False)
        self.channel_visibility_changed.emit()

    def on_automatic_color(self, cmap_name = None):
        cmap_name = str(self.combo_cmap.currentText())
        n = np.sum(self.selected)
        if n==0: return
        cmap = matplotlib.cm.get_cmap(cmap_name , n)

        self.viewer.by_channel_params.blockSignals(True)
        for i, c in enumerate(np.nonzero(self.selected)[0]):
            color = [ int(e*255) for e in  matplotlib.colors.ColorConverter().to_rgb(cmap(i)) ]
            self.viewer.by_channel_params['ch{}'.format(c), 'color'] = color
        self.viewer.all_params.sigTreeStateChanged.connect(self.viewer.on_param_change)
        self.viewer.by_channel_params.blockSignals(False)
        self.channel_color_changed.emit()

    def on_channel_visibility_changed(self):
        self.viewer.refresh()

    def on_channel_color_changed(self):
        self.viewer.refresh()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from .sourcebase import BaseDataSource
import sys


import numpy as np

try:
    import av
    HAVE_AV = True
except ImportError:
    HAVE_AV = False


AV_TIME_BASE = 1000000

def pts_to_frame(pts, time_base, frame_rate, start_time):
    return int(pts * time_base * frame_rate) - int(start_time * time_base * frame_rate)

def get_frame_rate(stream):

    if stream.average_rate.denominator and stream.average_rate.numerator:
        return float(stream.average_rate)
    if stream.time_base.denominator and stream.time_base.numerator:
        return 1.0/float(stream.time_base)
    else:
        raise ValueError("Unable to determine FPS")

#~ def get_frame_count(f, stream):

    #~ if stream.frames:
        #~ return stream.frames
    #~ elif stream.duration:
        #~ return pts_to_frame(stream.duration, float(stream.time_base), get_frame_rate(stream), 0)
    #~ elif f.duration:
        #~ return pts_to_frame(f.duration, 1/float(AV_TIME_BASE), get_frame_rate(stream), 0)

    #~ else:
        #~ raise ValueError("Unable to determine number for frames")



class FrameGrabber:
    """

    this is taken from pyav example here but without QT stuff.
    https://github.com/mikeboers/PyAV/blob/master/scratchpad/frame_seek_example.py
    """

    def __init__(self):
        self.file = None
        self.stream = None
        self.frame = None

        self.start_time = 0
        self.pts_seen = False
        self.nb_frames = None

        self.rate = None
        self.time_base = None

        self.last_frame = None
        self.last_frame_index = None

    def next_frame(self):

        frame_index = None

        rate = self.rate
        time_base = self.time_base

        self.pts_seen = False

        for packet in self.file.demux(self.stream):
            #~ print("    pkt", packet.pts, packet.dts, packet)
            if packet.pts:
                self.pts_seen = True

            for frame in packet.decode():
                #~ print('   frame', frame)
                if frame_index is None:

                    if self.pts_seen:
                        pts = frame.pts
                    else:
                        pts = frame.dts
                    #~ print('  pts',pts)
                    if not pts is None:
                        frame_index = pts_to_frame(pts, time_base, rate, self.start_time)

                elif not frame_index is None:
                    frame_index += 1


                yield frame_index, frame

    def get_frame(self, target_frame):
        #~ print('get_frame', target_frame)

        if target_frame == self.last_frame_index:
            frame = self.last_frame
        elif target_frame < 0 or target_frame >= self.nb_frames:
            frame = None
        elif self.last_frame_index is None or \
                (target_frame < self.last_frame_index) or \
                (target_frame > self.last_frame_index + 300):
            frame = self.get_frame_absolut_seek(target_frame)
        else:
            frame = None
            for i, (frame_index, next_frame) in enumerate(self.next_frame()):
                #~ print("   ", i, "NEXT at frame", next_frame, "at ts:", next_frame.pts,next_frame.dts)
                if frame_index is None or frame_index >= target_frame:
                    frame = next_frame
                    break

        if frame is not None:
            self.last_frame = frame
            self.last_frame_index = target_frame

        return frame



    def get_frame_absolut_seek(self, target_frame):
        #~ print('get_frame_absolut_seek', target_frame)
        #~ print('self.active_frame', self.active_frame)

        #~ if target_frame != self.active_frame:
            #~ print('YEP')
            #~ return
        #~ print 'seeking to', target_frame

        seek_frame = target_frame

        rate = self.rate
        time_base = self.time_base
        #~ print 'ici', rate, time_base, 'target_frame', target_frame

        frame = None
        reseek = 250

        original_target_frame_pts = None

        while reseek >= 0:

            # convert seek_frame to pts
            target_sec = seek_frame * 1/rate
            target_pts = int(target_sec / time_base) + self.start_time

            #~ print 'la', 'target_sec', target_sec, 'target_pts', target_pts

            if original_target_frame_pts is None:
                original_target_frame_pts = target_pts

            try:
                # PyAV >= 6.1.0
                self.file.seek(int(target_pts), stream=self.stream)
            except TypeError:
                # PyAV < 6.1.0
                self.stream.seek(int(target_pts))

            frame_index = None

            frame_cache = []

            for i, (frame_index, frame) in enumerate(self.next_frame()):

                #~ # optimization if the time slider has changed, the requested frame no longer valid
                #~ if target_frame != self.active_frame:
                    #~ print('YEP0 target_frame != self.active_frame', target_frame, self.active_frame)
                    #~ return

                #~ print("   ", i, "at frame", frame_index, "at ts:", frame.pts,frame.dts,"target:", target_pts, 'orig', original_target_frame_pts)

                if frame_index is None:
                    pass

                elif frame_index >= target_frame:
                    break

                frame_cache.append(frame)

            # Check if we over seeked, if we over seekd we need to seek to a earlier time
            # but still looking for the target frame
            if frame_index != target_frame:

                if frame_index is None:
                    over_seek = '?'
                else:
                    over_seek = frame_index - target_frame
                    if frame_index > target_frame:

                        #~ print over_seek, frame_cache
                        if over_seek <= len(frame_cache):
                            #~ print "over seeked by %i, using cache" % over_seek
                            frame = frame_cache[-over_seek]
                            break


                seek_frame -= 1
                reseek -= 1
                #~ print "over seeked by %s, backtracking.. seeking: %i target: %i retry: %i" % (str(over_seek),  seek_frame, target_frame, reseek)

            else:
                break

        #~ print('ici frame', frame)
        if reseek < 0:
            #~ print('YEP reseek < 0')
            #~ raise ValueError("seeking failed %i" % frame_index)
            return None

        # frame at this point should be the correct frame

        if frame:

            return frame

        else:
            return None
            #~ raise ValueError("seeking failed %i" % target_frame)

    def get_frame_count(self):

        frame_count = None

        if self.stream.frames:
            frame_count = self.stream.frames
        elif self.stream.duration:
            frame_count =  pts_to_frame(self.stream.duration, float(self.stream.time_base), get_frame_rate(self.stream), 0)
        elif self.file.duration:
            frame_count = pts_to_frame(self.file.duration, 1/float(AV_TIME_BASE), get_frame_rate(self.stream), 0)
        else:
            raise ValueError("Unable to determine number for frames")

        seek_frame = frame_count

        retry = 100

        while retry:
            target_sec = seek_frame * 1/ self.rate
            target_pts = int(target_sec / self.time_base) + self.start_time

            try:
                # PyAV >= 6.1.0
                self.file.seek(int(target_pts), stream=self.stream)
            except TypeError:
                # PyAV < 6.1.0
                self.stream.seek(int(target_pts))

            frame_index = None

            for frame_index, frame in self.next_frame():
                #~ print frame_index, frame
                continue

            if not frame_index is None:
                break
            else:
                seek_frame -= 1
                retry -= 1


        #~ print "frame count seeked", frame_index, "container frame count", frame_count

        return frame_index or frame_count

    def set_file(self, path):
        #~ print(path, type(path))
        self.file = av.open(path)
        #~ for s in self.file.streams:
            #~ print(s.type)
        #~ self.stream = next(s for s in self.file.streams if s.type == b'video') #py2
        self.stream = next(s for s in self.file.streams if s.type == 'video')

        self.rate = get_frame_rate(self.stream)
        self.time_base = float(self.stream.time_base)


        index, first_frame = next(self.next_frame())

        try:
            # PyAV >= 6.1.0
            self.file.seek(self.stream.start_time or 0, stream=self.stream)
        except TypeError:
            # PyAV < 6.1.0
            self.stream.seek(self.stream.start_time or 0)

        # find the pts of the first frame
        index, first_frame = next(self.next_frame())

        if self.pts_seen:
            pts = first_frame.pts
        else:
            pts = first_frame.dts

        self.start_time = pts or first_frame.dts

        #~ print("First pts", pts, self.stream.start_time, first_frame)

        #self.nb_frames = get_frame_count(self.file, self.stream)
        self.nb_frames = self.get_frame_count()




class MultiVideoFileSource( BaseDataSource):
    type = 'video'
    def __init__(self, video_filenames, video_times=None):
        assert HAVE_AV, 'PyAv is not installed'

        self.video_filenames = video_filenames
        self.video_times = video_times
        n = len(self.video_filenames)

        self.t_starts, self.t_stops, self.rates = [], [], []
        self.nb_frames = []
        self.frame_grabbers = []
        for video_filename in self.video_filenames:
            fg = FrameGrabber()
            fg.set_file(video_filename)
            self.frame_grabbers.append(fg)

            self.nb_frames.append(fg.get_frame_count())
            self.rates.append(fg.rate)
            self.t_starts.append(fg.start_time)
            self.t_stops.append(fg.start_time+fg.stream.duration*fg.time_base)

        self._t_start = min(self.t_starts)
        self._t_stop = max(self.t_stops)

    @property
    def nb_channel(self):
        return len(self.video_filenames)

    def get_channel_name(self, chan=0):
        return 'video {}'.format(chan)

    @property
    def t_start(self):
        return self._t_start

    @property
    def t_stop(self):
        return self._t_stop

    def time_to_frame_index(self, i, t):

        # if t is between frames, both methods return
        # the index of the frame *preceding* t
        if self.video_times is None:
            frame_index = int((t-self.t_starts[i])*self.rates[i])
        else:
            frame_index = np.searchsorted(self.video_times[i], t, side='right') - 1

        return frame_index



import sys


if sys.version_info > (3,):
    # py3x Qt5
    from . import  icons_py3 as icons
else:
    # py27 Qt4
    from . import  icons_py2_Qt4 as icons


# TODO : make py3_Qt4 (normally never the case)




import ephyviewer
import ephyviewer.base



def test_base():
    app = ephyviewer.mkQApp()
    win = ephyviewer.base.ViewerBase()
    win.show()
    app.exec_()




if __name__=='__main__':
    test_base()



import ephyviewer


from  ephyviewer.tests.testing_tools import make_fake_signals


def test_traceviewer():
    source = make_fake_signals()


    app = ephyviewer.mkQApp()
    view = ephyviewer.TimeFreqViewer(source=source, name='timefreq')
    #~ view.refresh()

    #~ for c in range(source.nb_channel):
        #~ view.by_channel_params['ch'+str(c), 'visible'] = True

    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)
    win.show()

    app.exec_()


if __name__=='__main__':
    test_traceviewer()




import ephyviewer
import numpy as np
import os

from  ephyviewer.tests.testing_tools import make_video_file


def test_InMemoryAnalogSignalSource():

    signals = np.random.randn(1000000, 16)
    sample_rate = 10000.
    t_start = 0.

    source = ephyviewer.InMemoryAnalogSignalSource(signals, sample_rate, t_start)

    assert source.nb_channel==16
    assert source.get_shape() == signals.shape
    assert np.all(source.get_chunk(i_start=50, i_stop=55) == signals[50:55])






def test_VideoMultiFileSource():
    import av

    #~ video_filenames = ['video0.avi', 'video1.avi', 'video2.avi',]
    video_filenames = ['video0.avi',]
    for filename in video_filenames:
        if not os.path.exists(filename):
            make_video_file(filename)

    videotimes = None
    source = ephyviewer.MultiVideoFileSource(video_filenames, videotimes)
    assert source.t_start==0
    assert source.t_stop==10.


def test_InMemoryEventSource():
    ev_times = np.arange(0, 10., .5)
    ev_labels = np.array(['Event0 num {}'.format(i) for i in range(ev_times.size)], dtype='U')
    event0 = { 'time':ev_times, 'label':ev_labels, 'name':  'Event0' }

    ev_times = np.arange(-6, 8., 2.)
    ev_labels = np.array(['Event1 num {}'.format(i) for i in range(ev_times.size)], dtype='U')
    event1 = { 'time':ev_times, 'label':ev_labels, 'name':  'Event1' }

    all_events = [event0, event1]

    source = ephyviewer.InMemoryEventSource(all_events=all_events)

    assert source.t_start==-6.
    assert source.t_stop==9.5
    assert source.get_size(0)==20


def test_InMemoryEpochSource():
    ep_times = np.arange(0, 10., .5)
    ep_durations = np.ones(ep_times.shape) * .1
    ep_labels = np.array(['Epoch0 num {}'.format(i) for i in range(ep_times.size)], dtype='U')
    epoch0 = { 'time':ep_times, 'duration':ep_durations,'label':ep_labels, 'name':  'Epoch0' }

    ep_times = np.arange(-6, 8., 2.)
    ep_durations = np.ones(ep_times.shape) * .2
    ep_labels = np.array(['Epoch1 num {}'.format(i) for i in range(ep_times.size)], dtype='U')
    epoch1 = { 'time':ep_times, 'duration':ep_durations, 'label':ep_labels, 'name':  'Epoch1' }

    all_epochs = [epoch0, epoch1]

    source = ephyviewer.InMemoryEpochSource(all_epochs=all_epochs)

    assert source.t_start==-6.
    assert source.t_stop==9.6
    assert source.get_size(0)==20


def test_spikesource():
    sike_times = np.arange(0, 10., .5)
    spikes0 = { 'time':sike_times, 'name':  'Unit#0' }

    sike_times = np.arange(-6, 8., 2.)
    spikes1 = { 'time':sike_times, 'name':  'unit#1' }

    all_spikes = [spikes0, spikes1]

    source = ephyviewer.InMemorySpikeSource(all_spikes=all_spikes)

    assert source.t_start==-6.
    assert source.t_stop==9.5
    assert source.get_size(0)==20



def test_neo_rawio_sources():
    #TODO make autorun neo tdtrawio test before
    from neo.rawio.tdtrawio import TdtRawIO

    dirname = '/tmp/files_for_testing_neo/tdt/aep_05/'
    neorawio = TdtRawIO(dirname=dirname)
    neorawio.parse_header()
    print(neorawio)

    sources = ephyviewer.get_sources_from_neo_rawio(neorawio)
    #~ print(sources)

    for s in sources['signal']:
        print(s.t_start, s.nb_channel, s.sample_rate)
        print(s.get_chunk(i_start=0, i_stop=1024).shape)

    for s in sources['epoch']:
        print(s.t_start, s.nb_channel)
        #~ print(s.get_chunk(i_start=0, i_stop=1024).shape)
        print(s.get_chunk_by_time(chan=0,  t_start=None, t_stop=None))

    for s in sources['spike']:
        print(s.t_start, s.nb_channel)
        print(s.get_chunk_by_time(chan=0,  t_start=None, t_stop=None))
        #~ print(s.get_chunk(i_start=0, i_stop=1024).shape)

def test_neo_object_sources():

    from neo.test.generate_datasets import generate_one_simple_segment
    import neo

    neo_seg = generate_one_simple_segment(supported_objects=[neo.Segment, neo.AnalogSignal, neo.Event, neo.Epoch, neo.SpikeTrain])

    sources = ephyviewer.get_sources_from_neo_segment(neo_seg)


    for s in sources['signal']:
        print(s.t_start, s.nb_channel, s.sample_rate)
        print(s.get_chunk(i_start=0, i_stop=1024).shape)

    for s in sources['epoch']:
        print(s.t_start, s.nb_channel)
        print(s.get_chunk_by_time(chan=0,  t_start=0, t_stop=10.))

    for s in sources['event']:
        print(s.t_start, s.nb_channel)
        print(s.get_chunk_by_time(chan=0,  t_start=0, t_stop=10.))

    for s in sources['spike']:
        print(s.t_start, s.nb_channel)
        print(s.get_chunk_by_time(chan=0,  t_start=0., t_stop=10.))
        #~ print(s.get_chunk(i_start=0, i_stop=1024).shape)





if __name__=='__main__':
    #~ test_InMemoryAnalogSignalSource()
    #~ test_VideoMultiFileSource()
    #~ test_InMemoryEventSource()
    #~ test_InMemoryEpochSource()
    #~ test_spikesource()
    #~ test_neo_rawio_sources()
    test_neo_object_sources()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

#~ import matplotlib.cm
#~ import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg

from .base import BaseMultiChannelViewer, Base_MultiChannel_ParamController
from .datasource import InMemoryAnalogSignalSource, AnalogSignalSourceWithScatter, NeoAnalogSignalSource, AnalogSignalFromNeoRawIOSource
from .tools import mkCachedBrush





#todo remove this
import time
import threading



default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 3., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'ylim_max', 'type': 'float', 'value': 10.},
    {'name': 'ylim_min', 'type': 'float', 'value': -10.},
    {'name': 'scatter_size', 'type': 'float', 'value': 10.,  'limits': (0,np.inf)},
    {'name': 'scale_mode', 'type': 'list', 'value': 'real_scale',
        'values':['real_scale', 'same_for_all', 'by_channel'] },
    {'name': 'auto_scale_factor', 'type': 'float', 'value': 0.1, 'step': 0.01, 'limits': (0,np.inf)},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'label_fill_color', 'type': 'color', 'value': '#222222DD'},
    {'name': 'label_size', 'type': 'int', 'value': 8, 'limits': (1,np.inf)},
    {'name': 'display_labels', 'type': 'bool', 'value': False},
    {'name': 'display_offset', 'type': 'bool', 'value': False},
    {'name': 'antialias', 'type': 'bool', 'value': False},
    {'name': 'decimation_method', 'type': 'list', 'value': 'min_max', 'values': ['min_max', 'mean', 'pure_decimate',  ]},
    {'name': 'line_width', 'type': 'float', 'value': 1., 'limits': (0, np.inf)},
    ]

default_by_channel_params = [
    {'name': 'color', 'type': 'color', 'value': "55FF00"},
    {'name': 'gain', 'type': 'float', 'value': 1, 'step': 0.1, 'decimals': 8},
    {'name': 'offset', 'type': 'float', 'value': 0., 'step': 0.1},
    {'name': 'visible', 'type': 'bool', 'value': True},
    ]




#TODO use Base_MultiChannel_ParamController instead of Base_ParamController
#to avoid code duplication



class TraceViewer_ParamController(Base_MultiChannel_ParamController):

    def __init__(self, parent=None, viewer=None):
        Base_MultiChannel_ParamController.__init__(self, parent=parent, viewer=viewer, with_visible=True, with_color=True)

        # raw_gains and raw_offsets are distinguished from adjustable gains and
        # offsets associated with this viewer because it makes placement of the
        # baselines and labels very easy for both raw and in-memory sources
        if isinstance(self.viewer.source, AnalogSignalFromNeoRawIOSource):
            # use raw_gains and raw_offsets from the raw source
            self.raw_gains = self.viewer.source.get_gains()
            self.raw_offsets = self.viewer.source.get_offsets()
        else:
            # use 1 and 0 for in-memory sources, which have already been scaled
            # properly
            self.raw_gains = np.ones(self.viewer.source.nb_channel)
            self.raw_offsets = np.zeros(self.viewer.source.nb_channel)

        #TODO put this somewhere

        #~ v.addWidget(QT.QLabel(self.tr('<b>Gain zoom (mouse wheel on graph):</b>'),self))
        #~ h = QT.QHBoxLayout()
        #~ v.addLayout(h)
        #~ for label, factor in [('--', 1./5.), ('-', 1./1.1), ('+', 1.1), ('++', 5.),]:
            #~ but = QT.QPushButton(label)
            #~ but.factor = factor
            #~ but.clicked.connect(self.on_but_ygain_zoom)
            #~ h.addWidget(but)

        #~ self.ygain_factor = 1.

    @property
    def selected(self):
        selected = np.ones(self.viewer.source.nb_channel, dtype=bool)
        if self.viewer.source.nb_channel>1:
            selected[:] = False
            selected[[ind.row() for ind in self.qlist.selectedIndexes()]] = True
        return selected

    @property
    def visible_channels(self):
        visible = [self.viewer.by_channel_params['ch{}'.format(i), 'visible'] for i in range(self.source.nb_channel)]
        return np.array(visible, dtype='bool')

    @property
    def gains(self):
        gains = [self.viewer.by_channel_params['ch{}'.format(i), 'gain'] for i in range(self.source.nb_channel)]
        return np.array(gains)

    @gains.setter
    def gains(self, val):
        for c, v in enumerate(val):
            self.viewer.by_channel_params['ch{}'.format(c), 'gain'] = v

    @property
    def offsets(self):
        offsets = [self.viewer.by_channel_params['ch{}'.format(i), 'offset'] for i in range(self.source.nb_channel)]
        return np.array(offsets)

    @offsets.setter
    def offsets(self, val):
        for c, v in enumerate(val):
            self.viewer.by_channel_params['ch{}'.format(c), 'offset'] = v

    @property
    def total_gains(self):
        # compute_rescale sets adjustable gains and offsets such that
        #     data_curves = (chunk * raw_gains + raw_offsets) * gains + offsets
        #                 = chunk * (raw_gains * gains) + (raw_offsets * gains + offsets)
        #                 = chunk * total_gains + total_offsets
        return self.raw_gains * self.gains

    @property
    def total_offsets(self):
        # compute_rescale sets adjustable gains and offsets such that
        #     data_curves = (chunk * raw_gains + raw_offsets) * gains + offsets
        #                 = chunk * (raw_gains * gains) + (raw_offsets * gains + offsets)
        #                 = chunk * total_gains + total_offsets
        return (self.raw_offsets * self.gains) + self.offsets



    def estimate_median_mad(self):
        # Estimates are performed on real values for both raw and in-memory
        # sources, i.e., on sigs = chunk * raw_gains + raw_offsets, where
        # raw_gains = 1 and raw_offsets = 0 for in-memory sources.

        #~ print('estimate_median_mad')
        #~ t0 = time.perf_counter()
        sigs = self.viewer.last_sigs_chunk
        assert sigs is not None, 'Need to debug this'
        #~ print(sigs)
        #~ print(sigs.shape)

        if sigs.shape[0]>1000:
            # to fast auto scale on long signal
            ind = np.random.randint(0, sigs.shape[0], size=1000)
            sigs = sigs[ind, :]

        if sigs.shape[0] > 0:
            sigs = sigs * self.raw_gains + self.raw_offsets  # calculate on real values
            self.signals_med = med = np.median(sigs, axis=0)
            self.signals_mad = np.median(np.abs(sigs-med),axis=0)*1.4826
            self.signals_min = np.min(sigs, axis=0)
            self.signals_max = np.max(sigs, axis=0)
        else:
            # dummy median/mad/...
            n = self.viewer.source.nb_channel
            self.signals_med = np.zeros(n)
            self.signals_mad = np.ones(n)
            self.signals_min = -np.ones(n)
            self.signals_max = np.ones(n)

        #~ t1 = time.perf_counter()
        #~ print('estimate_median_mad DONE', t1-t0)
        #~ print('self.signals_med', self.signals_med)

    def compute_rescale(self):
        # estimate_median_mad operates on real values, i.e., on
        # sigs = chunk * raw_gains + raw_offsets. Consequently, the gains and
        # offsets computed here will map real values to plot coords:
        #     data_curves = (chunk * raw_gains + raw_offsets) * gains + offsets
        #                 = chunk * (raw_gains * gains) + (raw_offsets * gains + offsets)
        #                 = chunk * total_gains + total_offsets

        scale_mode = self.viewer.params['scale_mode']
        #~ print('compute_rescale', scale_mode)

        self.viewer.all_params.blockSignals(True)

        gains = np.ones(self.viewer.source.nb_channel)
        offsets = np.zeros(self.viewer.source.nb_channel)
        nb_visible = np.sum(self.visible_channels)
        #~ self.ygain_factor = 1
        if self.viewer.last_sigs_chunk is not None and self.viewer.last_sigs_chunk is not []:
            self.estimate_median_mad()

            if scale_mode=='real_scale':
                self.viewer.params['ylim_min'] = np.nanmin(self.signals_min[self.visible_channels])
                self.viewer.params['ylim_max'] = np.nanmax(self.signals_max[self.visible_channels])
            else:
                if scale_mode=='same_for_all':
                    gains[self.visible_channels] = np.ones(nb_visible, dtype=float) / max(self.signals_mad[self.visible_channels]) * self.viewer.params['auto_scale_factor']
                elif scale_mode=='by_channel':
                    gains[self.visible_channels] = np.ones(nb_visible, dtype=float) / self.signals_mad[self.visible_channels] * self.viewer.params['auto_scale_factor']
                offsets[self.visible_channels] = np.arange(nb_visible)[::-1] - self.signals_med[self.visible_channels]*gains[self.visible_channels]
                self.viewer.params['ylim_min'] = -0.5
                self.viewer.params['ylim_max'] = nb_visible-0.5

        self.gains = gains
        self.offsets = offsets
        self.viewer.all_params.blockSignals(False)

    def on_channel_visibility_changed(self):
        #~ print('on_channel_visibility_changed')
        self.compute_rescale()
        self.viewer.refresh()

    def on_but_ygain_zoom(self):
        factor = self.sender().factor
        self.apply_ygain_zoom(factor)

    def apply_ygain_zoom(self, factor_ratio, chan_index=None):

        scale_mode = self.viewer.params['scale_mode']

        self.viewer.all_params.blockSignals(True)
        if scale_mode=='real_scale':
            #~ self.ygain_factor *= factor_ratio

            self.viewer.params['ylim_max'] = self.viewer.params['ylim_max']*factor_ratio
            self.viewer.params['ylim_min'] = self.viewer.params['ylim_min']*factor_ratio

            pass
            #TODO ylims
        else :
            #~ self.ygain_factor *= factor_ratio
            if not hasattr(self, 'signals_med'):
                self.estimate_median_mad()
            if scale_mode=='by_channel' and chan_index is not None:
                # factor_ratio should be applied to only the desired channel,
                # so turn the scalar factor into a vector of ones everywhere
                # except at chan_index
                factor_ratio_vector = np.ones(self.source.nb_channel)
                factor_ratio_vector[chan_index] = factor_ratio
                factor_ratio = factor_ratio_vector
            self.offsets = self.offsets + self.signals_med*self.gains * (1-factor_ratio)
            self.gains = self.gains * factor_ratio

        self.viewer.all_params.blockSignals(False)

        self.viewer.refresh()
        #~ print('apply_ygain_zoom', factor_ratio)#, 'self.ygain_factor', self.ygain_factor)

    def apply_label_drag(self, label_y, chan_index):
        self.viewer.by_channel_params['ch{}'.format(chan_index), 'offset'] = label_y - self.signals_med[chan_index]*self.gains[chan_index]




class DataGrabber(QT.QObject):
    data_ready = QT.pyqtSignal(float, float, float, object, object, object, object, object)

    def __init__(self, source, viewer, parent=None):
        QT.QObject.__init__(self, parent)
        self.source = source
        self.viewer = viewer
        self._max_point = 3000

    def get_data(self, t, t_start, t_stop, total_gains, total_offsets, visibles, decimation_method):

        i_start, i_stop = self.source.time_to_index(t_start), self.source.time_to_index(t_stop) + 2
        #~ print(t_start, t_stop, i_start, i_stop)

        ds_ratio = (i_stop - i_start)//self._max_point + 1
        #~ print()
        #~ print('ds_ratio', ds_ratio, 'i_start i_stop', i_start, i_stop  )

        if ds_ratio>1:
            i_start = i_start - (i_start%ds_ratio)
            i_stop = i_stop - (i_stop%ds_ratio)
            #~ print('i_start, i_stop', i_start, i_stop)

        #clip it
        i_start = max(0, i_start)
        i_start = min(i_start, self.source.get_length())
        i_stop = max(0, i_stop)
        i_stop = min(i_stop, self.source.get_length())
        if ds_ratio>1:
            #after clip
            i_start = i_start - (i_start%ds_ratio)
            i_stop = i_stop - (i_stop%ds_ratio)

        #~ print('final i_start i_stop', i_start, i_stop  )

        sigs_chunk = self.source.get_chunk(i_start=i_start, i_stop=i_stop)



        #~ print('sigs_chunk.shape', sigs_chunk.shape)
        data_curves = sigs_chunk[:, visibles].T.copy()
        if data_curves.dtype!='float32':
            data_curves = data_curves.astype('float32')

        if ds_ratio>1:


            small_size = (data_curves.shape[1]//ds_ratio)
            if decimation_method == 'min_max':
                small_size *= 2

            small_arr = np.empty((data_curves.shape[0], small_size), dtype=data_curves.dtype)

            if decimation_method == 'min_max' and data_curves.size>0:
                full_arr = data_curves.reshape(data_curves.shape[0], -1, ds_ratio)
                small_arr[:, ::2] = full_arr.max(axis=2)
                small_arr[:, 1::2] = full_arr.min(axis=2)
            elif decimation_method == 'mean' and data_curves.size>0:
                full_arr = data_curves.reshape(data_curves.shape[0], -1, ds_ratio)
                small_arr[:, :] = full_arr.mean(axis=2)
            elif decimation_method == 'pure_decimate':
                small_arr[:, :] = data_curves[:, ::ds_ratio]
            elif data_curves.size == 0:
                pass


            data_curves = small_arr

        #~ print(data_curves.shape)

        data_curves *= total_gains[visibles, None]
        data_curves += total_offsets[visibles, None]
        dict_curves ={}
        for i, c in enumerate(visibles):
            dict_curves[c] = data_curves[i, :]

        #~ print(ds_ratio)
        t_start2 = self.source.index_to_time(i_start)
        times_curves = np.arange(data_curves.shape[1], dtype='float64') # ensure high temporal precision (see issue #28)
        times_curves /= self.source.sample_rate/ds_ratio
        if ds_ratio>1 and decimation_method == 'min_max':
            times_curves /=2
        times_curves += t_start2

        dict_scatter = None
        if self.source.with_scatter:
            pass
            dict_scatter = {}
            for k in self.source.get_scatter_babels():
                x, y = [[]], [[]]
                for i, c in enumerate(visibles):
                    scatter_inds = self.source.get_scatter(i_start=i_start, i_stop=i_stop, chan=c, label=k)
                    if scatter_inds is None: continue
                    x.append((scatter_inds-i_start)/self.source.sample_rate+t_start2)
                    y.append(sigs_chunk[scatter_inds-i_start, c]*total_gains[c]+total_offsets[c])

                dict_scatter[k] = (np.concatenate(x), np.concatenate(y))

        return t, t_start, t_stop, visibles, dict_curves, times_curves, sigs_chunk, dict_scatter

    def on_request_data(self, t, t_start, t_stop, total_gains, total_offsets, visibles, decimation_method):
        #~ print('on_request_data', t_start, t_stop)

        if self.viewer.t != t:
            #~ print('on_request_data not same t')
            return

        t, t_start, t_stop, visibles, dict_curves, times_curves,\
                    sigs_chunk, dict_scatter = self.get_data(t, t_start, t_stop, total_gains, total_offsets, visibles, decimation_method)


        #~ print('on_request_data', threading.get_ident())
        #~ time.sleep(1.)
        self.data_ready.emit(t, t_start, t_stop, visibles, dict_curves, times_curves, sigs_chunk, dict_scatter)



class TraceLabelItem(pg.TextItem):

    label_dragged = QT.pyqtSignal(float)
    label_ygain_zoom = QT.pyqtSignal(float)

    def __init__(self, **kwargs):
        pg.TextItem.__init__(self, **kwargs)

        self.dragOffset = None

    def mouseDragEvent(self, ev):
        '''Emit the new y-coord of the label as it is dragged'''

        if ev.button() != QT.LeftButton:
            ev.ignore()
            return
        else:
            ev.accept()

        if ev.isStart():
            # To avoid snapping the label to the mouse cursor when the drag
            # starts, we determine the offset of the position where the button
            # was first pressed down relative to the label's origin/anchor, in
            # plot coordinates
            self.dragOffset = self.mapToParent(ev.buttonDownPos()) - self.pos()

        # The new y-coord for the label is the mouse's current position during
        # the drag with the initial offset removed
        new_y = (self.mapToParent(ev.pos()) - self.dragOffset).y()
        self.label_dragged.emit(new_y)

    def wheelEvent(self, ev):
        '''Emit a yzoom factor for the associated trace'''
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 5. if ev.delta()>0 else 1/5.
        else:
            z = 1.1 if ev.delta()>0 else 1/1.1
        self.label_ygain_zoom.emit(z)
        ev.accept()


class TraceViewer(BaseMultiChannelViewer):
    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = TraceViewer_ParamController

    request_data = QT.pyqtSignal(float, float, float, object, object, object, object)

    def __init__(self, useOpenGL=None, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()

        # useOpenGL=True eliminates the extremely poor performance associated
        # with TraceViewer's line_width > 1.0, but it also degrades overall
        # performance somewhat and is reportedly unstable
        self.set_layout(useOpenGL=useOpenGL)

        self.make_param_controller()

        self.viewBox.doubleclicked.connect(self.show_params_controller)

        self.initialize_plot()

        self.last_sigs_chunk = None

        self.thread = QT.QThread(parent=self)
        self.datagrabber = DataGrabber(source=self.source, viewer=self)
        self.datagrabber.moveToThread(self.thread)
        self.thread.start()


        self.datagrabber.data_ready.connect(self.on_data_ready)
        self.request_data.connect(self.datagrabber.on_request_data)

        self.params.param('xsize').setLimits((0, np.inf))


    @classmethod
    def from_numpy(cls, sigs, sample_rate, t_start, name, channel_names=None,
                scatter_indexes=None, scatter_channels=None, scatter_colors=None):

        if scatter_indexes is None:
            source = InMemoryAnalogSignalSource(sigs, sample_rate, t_start, channel_names=channel_names)
        else:
            source = AnalogSignalSourceWithScatter(sigs, sample_rate, t_start, channel_names=channel_names,
                                            scatter_indexes=scatter_indexes, scatter_channels=scatter_channels, scatter_colors=scatter_colors)
        view = cls(source=source, name=name)

        return view

    @classmethod
    def from_neo_analogsignal(cls, neo_anasig, name):
        source = NeoAnalogSignalSource(neo_anasig)
        view = cls(source=source, name=name)
        return view

    def closeEvent(self, event):
        event.accept()
        self.thread.quit()
        self.thread.wait()

    def initialize_plot(self):

        self.vline = pg.InfiniteLine(angle = 90, movable = False, pen = self.params['vline_color'])
        self.vline.setZValue(1) # ensure vline is above plot elements
        self.plot.addItem(self.vline)

        self.curves = []
        self.channel_labels = []
        self.channel_offsets_line = []
        for c in range(self.source.nb_channel):
            color = self.by_channel_params['ch{}'.format(c), 'color']
            curve = pg.PlotCurveItem(pen='#7FFF00', downsampleMethod='peak', downsample=1,
                            autoDownsample=False, clipToView=True, antialias=False)#, connect='finite')
            self.plot.addItem(curve)
            self.curves.append(curve)

            ch_name = '{}: {}'.format(c, self.source.get_channel_name(chan=c))
            label = TraceLabelItem(text=ch_name, color=color, anchor=(0, 0.5), border=None, fill=self.params['label_fill_color'])
            label.setZValue(2) # ensure labels are drawn above scatter
            font = label.textItem.font()
            font.setPointSize(self.params['label_size'])
            label.setFont(font)
            label.label_dragged.connect(lambda label_y, chan_index=c: self.params_controller.apply_label_drag(label_y, chan_index))
            label.label_ygain_zoom.connect(lambda factor_ratio, chan_index=c: self.params_controller.apply_ygain_zoom(factor_ratio, chan_index))

            self.plot.addItem(label)
            self.channel_labels.append(label)

            offset_line = pg.InfiniteLine(angle = 0, movable = False, pen = '#7FFF00')
            self.plot.addItem(offset_line)
            self.channel_offsets_line.append(offset_line)

        if self.source.with_scatter:
            self.scatter = pg.ScatterPlotItem(size=self.params['scatter_size'], pxMode = True)
            self.plot.addItem(self.scatter)



        self.viewBox.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)
        self.viewBox.ygain_zoom.connect(self.params_controller.apply_ygain_zoom)

    def on_param_change(self, params=None, changes=None):
        #~ print('on_param_change')
        #track if new scale mode
        for param, change, data in changes:
            if change != 'value': continue
            if param.name()=='scale_mode':
                self.params_controller.compute_rescale()
            if param.name()=='antialias':
                for curve in self.curves:
                    curve.updateData(antialias=self.params['antialias'])
            if param.name()=='scatter_size':
                if self.source.with_scatter:
                    self.scatter.setSize(self.params['scatter_size'])
            if param.name()=='vline_color':
                self.vline.setPen(color=self.params['vline_color'])
            if param.name()=='label_fill_color':
                for label in self.channel_labels:
                    label.fill = pg.mkBrush(self.params['label_fill_color'])
            if param.name()=='label_size':
                for label in self.channel_labels:
                    font = label.textItem.font()
                    font.setPointSize(self.params['label_size'])
                    label.setFont(font)


        self.refresh()

    def auto_scale(self):
        #~ print('auto_scale', self.last_sigs_chunk)
        if self.last_sigs_chunk is None:
            xsize = self.params['xsize']
            xratio = self.params['xratio']
            t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
            visibles, = np.nonzero(self.params_controller.visible_channels)
            total_gains = self.params_controller.total_gains
            total_offsets = self.params_controller.total_offsets
            _, _, _, _, _, _,sigs_chunk, _ = self.datagrabber.get_data(self.t, t_start, t_stop, total_gains,
                                            total_offsets, visibles, self.params['decimation_method'])
            self.last_sigs_chunk = sigs_chunk

        self.params_controller.compute_rescale()
        self.refresh()

    def refresh(self):
        #~ print('TraceViewer.refresh', 't', self.t)
        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
        visibles, = np.nonzero(self.params_controller.visible_channels)
        total_gains = self.params_controller.total_gains
        total_offsets = self.params_controller.total_offsets

        self.request_data.emit(self.t, t_start, t_stop, total_gains, total_offsets, visibles, self.params['decimation_method'])


    def on_data_ready(self, t,   t_start, t_stop, visibles, dict_curves, times_curves, sigs_chunk, dict_scatter):
        #~ print('on_data_ready', t, t_start, t_stop)

        if self.t != t:
            #~ print('on_data_ready not same t')
            return

        self.graphicsview.setBackground(self.params['background_color'])

        self.last_sigs_chunk = sigs_chunk

        offsets = self.params_controller.offsets
        gains = self.params_controller.gains
        if not hasattr(self.params_controller, 'signals_med'):
            self.params_controller.estimate_median_mad()
        signals_med = self.params_controller.signals_med


        for i, c in enumerate(visibles):
            self.curves[c].show()
            self.curves[c].setData(times_curves, dict_curves[c])

            color = self.by_channel_params['ch{}'.format(c), 'color']
            self.curves[c].setPen(color=color, width=self.params['line_width'])

            if self.params['display_labels']:
                self.channel_labels[c].show()
                self.channel_labels[c].setPos(t_start, offsets[c] + signals_med[c]*gains[c])
                self.channel_labels[c].setColor(color)
            else:
                self.channel_labels[c].hide()

            if self.params['display_offset']:
                self.channel_offsets_line[c].show()
                self.channel_offsets_line[c].setPos(offsets[c])
                self.channel_offsets_line[c].setPen(color)
            else:
                self.channel_offsets_line[c].hide()

        for c in range(self.source.nb_channel):
            if c not in visibles:
                self.curves[c].hide()
                self.channel_labels[c].hide()
                self.channel_offsets_line[c].hide()

        if dict_scatter is not None:
            self.scatter.clear()
            all_x = []
            all_y = []
            all_brush = []
            for k, (x, y) in dict_scatter.items():
                all_x.append(x)
                all_y.append(y)

                # here we must use cached brushes to avoid issues with
                # the SymbolAtlas in pyqtgraph >= 0.11.1.
                # see https://github.com/NeuralEnsemble/ephyviewer/issues/132
                color = self.source.scatter_colors.get(k, '#FFFFFF')
                all_brush.append(np.array([mkCachedBrush(color)]*len(x)))

            if len(all_x):
                all_x = np.concatenate(all_x)
                all_y = np.concatenate(all_y)
                all_brush = np.concatenate(all_brush)
                self.scatter.setData(x=all_x, y=all_y, brush=all_brush)

        self.vline.setPos(self.t)
        self.plot.setXRange( t_start, t_stop, padding = 0.0)
        self.plot.setYRange(self.params['ylim_min'], self.params['ylim_max'], padding = 0.0)

        #~ self.graphicsview.repaint()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

import matplotlib.cm
import matplotlib.colors

from .myqt import QT
import pyqtgraph as pg

from .base import BaseMultiChannelViewer, Base_MultiChannel_ParamController
from .datasource import InMemoryEpochSource, NeoEpochSource


default_params = [
    {'name': 'xsize', 'type': 'float', 'value': 3., 'step': 0.1},
    {'name': 'xratio', 'type': 'float', 'value': 0.3, 'step': 0.1, 'limits': (0,1)},
    {'name': 'background_color', 'type': 'color', 'value': 'k'},
    {'name': 'vline_color', 'type': 'color', 'value': '#FFFFFFAA'},
    {'name': 'label_fill_color', 'type': 'color', 'value': '#222222DD'},
    {'name': 'label_size', 'type': 'int', 'value': 8, 'limits': (1,np.inf)},
    {'name': 'display_labels', 'type': 'bool', 'value': True},
    ]

default_by_channel_params = [
    {'name': 'color', 'type': 'color', 'value': "55FF00"},
    {'name': 'visible', 'type': 'bool', 'value': True},
    ]




class EpochViewer_ParamController(Base_MultiChannel_ParamController):
    pass




class RectItem(pg.GraphicsWidget):

    clicked = QT.pyqtSignal(int)
    doubleclicked = QT.pyqtSignal(int)

    def __init__(self, rect, border = 'r', fill = 'g', id = -1):
        pg.GraphicsWidget.__init__(self)
        self.rect = rect
        self.border= border
        self.fill= fill
        self.id = id

    def boundingRect(self):
        return QT.QRectF(0, 0, self.rect[2], self.rect[3])

    def paint(self, p, *args):
        p.setPen(pg.mkPen(self.border))
        p.setBrush(pg.mkBrush(self.fill))
        p.drawRect(self.boundingRect())

    def mouseClickEvent(self, event):
        if event.button()== QT.LeftButton:
            event.accept()
            if event.double():
                self.doubleclicked.emit(self.id)
            else:
                self.clicked.emit(self.id)
        else:
            event.ignore()


class DataGrabber(QT.QObject):
    data_ready = QT.pyqtSignal(float, float, object, object)

    def __init__(self, source, parent=None):
        QT.QObject.__init__(self, parent)
        self.source = source

    def on_request_data(self, t_start, t_stop, visibles):
        data = {}
        for e, chan in enumerate(visibles):
            data[chan] = self.source.get_chunk_by_time(chan=chan,  t_start=t_start, t_stop=t_stop)
        self.data_ready.emit(t_start, t_stop, visibles, data)


class EpochViewer(BaseMultiChannelViewer):
    _default_params = default_params
    _default_by_channel_params = default_by_channel_params

    _ControllerClass = EpochViewer_ParamController

    request_data = QT.pyqtSignal(float, float, object)

    def __init__(self, **kargs):
        BaseMultiChannelViewer.__init__(self, **kargs)

        self.make_params()
        self.set_layout()
        self.make_param_controller()
        self.viewBox.doubleclicked.connect(self.show_params_controller)
        self.initialize_plot()

        self.thread = QT.QThread(parent=self)
        self.datagrabber = DataGrabber(source=self.source)
        self.datagrabber.moveToThread(self.thread)
        self.thread.start()


        self.datagrabber.data_ready.connect(self.on_data_ready)
        self.request_data.connect(self.datagrabber.on_request_data)

        self.params.param('xsize').setLimits((0, np.inf))

    @classmethod
    def from_numpy(cls, all_epochs, name):
        source = InMemoryEpochSource(all_epochs)
        view = cls(source=source, name=name)
        return view

    @classmethod
    def from_neo_epochs(cls, neo_epochs, name):
        source = NeoEpochSource(neo_epochs)
        view = cls(source=source, name=name)
        return view

    def closeEvent(self, event):
        event.accept()
        self.thread.quit()
        self.thread.wait()


    def initialize_plot(self):
        self.viewBox.xsize_zoom.connect(self.params_controller.apply_xsize_zoom)

    def refresh(self):
        xsize = self.params['xsize']
        xratio = self.params['xratio']
        t_start, t_stop = self.t-xsize*xratio , self.t+xsize*(1-xratio)
        visibles, = np.nonzero(self.params_controller.visible_channels)
        self.request_data.emit(t_start, t_stop, visibles)

    def on_data_ready(self, t_start, t_stop, visibles, data):
        self.plot.clear()
        self.graphicsview.setBackground(self.params['background_color'])

        for e, chan in enumerate(visibles):

            if len(data[chan])==3:
                times, durations, labels = data[chan]
            elif len(data[chan])==4:
                times, durations, labels, _ = data[chan]
            else:
                raise ValueError("data has unexpected dimensions")

            color = self.by_channel_params.children()[chan].param('color').value()
            color2 = QT.QColor(color)
            color2.setAlpha(130)

            ypos = visibles.size-e-1

            for i in range(times.size):
                item = RectItem([times[i],  ypos,durations[i], .9],  border = color, fill = color2)
                item.setPos(times[i],  visibles.size-e-1)
                self.plot.addItem(item)

            if self.params['display_labels']:
                label_name = '{}: {}'.format(chan, self.source.get_channel_name(chan=chan))
                label = pg.TextItem(label_name, color=color, anchor=(0, 0.5), border=None, fill=self.params['label_fill_color'])
                font = label.textItem.font()
                font.setPointSize(self.params['label_size'])
                label.setFont(font)
                self.plot.addItem(label)
                label.setPos(t_start, ypos+0.45)

        self.vline = pg.InfiniteLine(angle = 90, movable = False, pen = self.params['vline_color'])
        self.vline.setZValue(1) # ensure vline is above plot elements
        self.plot.addItem(self.vline)

        self.vline.setPos(self.t)
        self.plot.setXRange( t_start, t_stop, padding = 0.0)
        self.plot.setYRange( 0, visibles.size)




import ephyviewer
import numpy as np
import os



def test_neoviewer():
    #TODO make autorun neo tdtrawio test before
    from neo.rawio.tdtrawio import TdtRawIO

    dirname = '/tmp/files_for_testing_neo/tdt/aep_05/'
    neorawio = TdtRawIO(dirname=dirname)
    neorawio.parse_header()
    print(neorawio)

    sources = ephyviewer.get_sources_from_neo_rawio(neorawio)


    app = ephyviewer.mkQApp()
    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)

    for i, sig_source in enumerate(sources['signal']):
        view = ephyviewer.TraceViewer(source=sig_source, name='signal {}'.format(i))
        win.add_view(view)

    for i, ep_source in enumerate(sources['epoch']):
        view = ephyviewer.EpochViewer(source=ep_source, name='epochs')
        win.add_view(view)

    win.show()

    app.exec_()



if __name__=='__main__':
    test_neoviewer()



from setuptools import setup
import os


install_requires = [
                    'numpy',
                    #~ 'PyQt5',
                    'pyqtgraph>=0.10.0',
                    'matplotlib>=2.0',
                    'scipy',
                    ]

# Read in the README to serve as the long_description, which will be presented
# on pypi.org as the project description.
with open('README.rst', 'r') as f:
    long_description = f.read()

with open('ephyviewer/version.py') as f:
    d = {}
    exec(f.read(), None, d)
    version = d['version']


entry_points={'console_scripts': ['ephyviewer=ephyviewer.scripts:launch_standalone_ephyviewer']}


setup(
    name = 'ephyviewer',
    version=version,
    packages = ['ephyviewer', 'ephyviewer.datasource', 'ephyviewer.tests', 'ephyviewer.icons'],
    install_requires=install_requires,
    author = 'S.Garcia, Jeffrey Gill',
    author_email = '',  # left blank because multiple emails cannot be provided
    description = 'Simple viewers for ephys signals, events, video and more',
    entry_points = entry_points,
    long_description = long_description,
    license = 'MIT',
    url='https://github.com/NeuralEnsemble/ephyviewer',
    project_urls={
        'Documentation': 'https://ephyviewer.readthedocs.io/en/latest/',
        'Source code': 'https://github.com/NeuralEnsemble/ephyviewer/',
        'Bug tracker': 'https://github.com/NeuralEnsemble/ephyviewer/issues',
        'Change log': 'https://ephyviewer.readthedocs.io/en/latest/releasenotes.html',
    },
    classifiers = [
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        ]
)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

from .sourcebase import BaseDataSource

import matplotlib.pyplot as plt
import matplotlib.colors


class BaseAnalogSignalSource(BaseDataSource):
    type = 'AnalogSignal'

    def get_length(self):
        raise(NotImplementedError)

    def get_shape(self):
        return (self.get_length(), self.nb_channel)

    def get_chunk(self, i_start=None, i_stop=None):
        raise(NotImplementedError)

    def time_to_index(self, t):
        return int((t-self.t_start)*self.sample_rate)

    def index_to_time(self, ind):
        return float(ind/self.sample_rate) + self.t_start


class InMemoryAnalogSignalSource(BaseAnalogSignalSource):
    def __init__(self, signals, sample_rate, t_start, channel_names=None):
        BaseAnalogSignalSource.__init__(self)

        self.signals = signals
        self.sample_rate = float(sample_rate)
        self._t_start = float(t_start)
        self._t_stop = self.signals.shape[0]/self.sample_rate + float(t_start)
        self.channel_names = channel_names
        if channel_names is None:
            self.channel_names = ['Channel {:3}'.format(c) for c in range(self.signals.shape[1])]

        self.with_scatter = False

    @property
    def nb_channel(self):
        return self.signals.shape[1]

    def get_channel_name(self, chan=0):
        return self.channel_names[chan]

    @property
    def t_start(self):
        return self._t_start

    @property
    def t_stop(self):
        return self._t_stop

    def get_length(self):
        return self.signals.shape[0]

    def get_chunk(self, i_start=None, i_stop=None):
        return self.signals[i_start:i_stop, :]




class AnalogSignalSourceWithScatter(InMemoryAnalogSignalSource):
    def __init__(self, signals, sample_rate, t_start, scatter_indexes, scatter_channels, scatter_colors=None, channel_names=None):
        InMemoryAnalogSignalSource.__init__(self, signals, sample_rate, t_start, channel_names=channel_names)
        self.with_scatter = True

        #todo test and assert self.scatter_indexes sorted for eack k
        self.scatter_indexes = scatter_indexes
        self.scatter_channels = scatter_channels
        self.scatter_colors = scatter_colors

        self._labels = list(self.scatter_indexes.keys())

        if self.scatter_colors is None:
            self.scatter_colors = {}
            n = len(self._labels)
            colors = plt.cm.get_cmap('Accent', n)
            for i,k in enumerate(self._labels):
                self.scatter_colors[k] = matplotlib.colors.to_hex(colors(i))

    def get_scatter_babels(self):
        return self._labels

    def get_scatter(self,  i_start=None, i_stop=None, chan=None, label=None):
        if chan not in self.scatter_channels[label]:
            return None

        inds = self.scatter_indexes[label]
        i1 = np.searchsorted(inds, i_start, side='left')
        i2 = np.searchsorted(inds, i_stop, side='left')
        return inds[i1:i2]
